# SPDX-License-Identifier: MIT

import struct

from enum import IntEnum
from ..utils import *
from ..malloc import Heap

__all__ = ["DART8110Regs", "DART8110"]

class R_PARAMS_0(Register32):
    CLIENT_PARTITIONS_SUPPORTED = 29
    LOG2_PGSZ = 27, 24
    LOG2_TE_COUNT = 22, 20
    TLB_SET_COUNT = 11, 0

class R_PARAMS_4(Register32):
    LOG2_NUM_WAYS = 30, 28
    NUM_ASCS = 25, 24
    NUM_W_PORTS = 22, 20
    NUM_R_PORTS = 18, 16
    NUM_APFS = 15, 8
    SUPPORT_STT_PREFETCH = 6
    SUPPORT_TLB_PREFETCH = 5
    SUPPORT_CTC_PREFETCH = 4
    SUPPORT_HW_FLUSH = 3
    SUPPORT_TZ_TAGGER = 2
    SUPPORT_REG_LOCK = 1
    SUPPORT_FULL_BYPASS = 0

class R_PARAMS_8(Register32):
    PA_WIDTH = 29, 24
    VA_WIDTH = 21, 16
    VERS_MAJ = 15, 8
    VERS_MIN = 7, 0

class R_PARAMS_C(Register32):
    NUM_CLIENTS = 24, 16
    NUM_SIDS = 8, 0

class R_ERROR(Register32):
    FLAG = 31
    SMMU = 30
    REGION_PROTECT = 29
    WRITE_nREAD = 28
    SID = 27, 20
    SECONDARY = 19
    FILL_REGION = 18
    BPF_REJECT = 14
    EXTERNAL = 13
    STT_FLUSH = 12
    STT_MISMATCH = 11
    APF_REJECT = 10
    DROP_PROTECT = 9
    CTRR_WRITE_PROTECT = 8
    AXI_ERROR = 7
    AXI_DECODE = 6
    READ_FAULT = 5
    WRITE_FAULT = 4
    NO_PTE = 3
    NO_PMD = 2  # "STE"
    NO_PGD = 1  # "CTE"
    NO_TTBR = 0

class R_TLB_OP(Register32):
    BUSY = 31

    # None of these bits are supported on hwrev 1
    HARDWARE_FLUSH = 30
    FLUSH_VA_RANGE = 14
    ENABLE_STT_FLUSH = 13
    DISABLE_STC_FLUSH = 12

    # 0 = flush all
    # 1 = flush SID
    # 2 = TLB read
    # 3 = TLB write????
    # 4 = flush unlock, definitely not supported on hwrev 1
    OP = 10, 8
    STREAM = 7, 0

class R_TLB_OP_IDX(Register32):
    SET = 13, 8
    WAY = 6, 4
    TE = 2, 0

class R_PROTECT(Register32):
    LOCK_TZ_SELECT = 4
    LOCK_TZ_CONFIG = 3
    # This bit can be set, but unknown what it protects
    _BIT2 = 2
    LOCK_REG_4xx = 1
    LOCK_TCR_TTBR = 0

class R_DIAG_LOCK(Register32):
    # FIXME: how does this work exactly?
    LOCK_ON_ERR = 1
    LOCK = 0

class R_TCR(Register32):
    REMAP = 11, 8
    REMAP_EN = 7
    FOUR_LEVELS = 3     # not supported on hwrev 1
    BYPASS_DAPF = 2
    BYPASS_DART = 1
    TRANSLATE_ENABLE = 0

class R_TTBR(Register32):
    ADDR = 29, 2
    VALID = 0

class PTE(Register64):
    SP_START = 63, 52
    SP_END = 51, 40
    OFFSET = 37, 10
    RDPROT = 3
    WRPROT = 2
    UNCACHABLE = 1
    VALID = 0

class DART8110Regs(RegMap):
    PARAMS_0 = 0x000, R_PARAMS_0
    PARAMS_4 = 0x004, R_PARAMS_4
    PARAMS_8 = 0x008, R_PARAMS_8
    PARAMS_C = 0x00C, R_PARAMS_C
    # Unknown RO
    REG_0x10 = 0x010, Register32
    REG_0x14 = 0x014, Register32    # hwrev 2 only

    TLB_OP      = 0x080, R_TLB_OP
    TLP_OP_IDX  = 0x084, R_TLB_OP_IDX
    TLB_TAG_LO  = 0x088, Register32
    TLB_TAG_HI  = 0x08c, Register32    # hwrev 2 only
    TLB_PA_LO   = 0x090, Register32
    TLB_PA_HI   = 0x094, Register32
    TLB_START_DVA_PAGE  = 0x098, Register32    # hwrev 2 only
    TLB_END_DVA_PAGE    = 0x0a0, Register32    # hwrev 2 only

    ERROR           = 0x100, R_ERROR
    ERROR_DISABLE   = 0x104, R_ERROR

    # Found via register bruteforcing
    STREAM_UNK_SET  = irange(0x120, 8, 4), Register32
    STREAM_UNK_CLR  = irange(0x140, 8, 4), Register32

    # these are all accessed by error interrupt handler
    REG_0x160   = 0x160, Register32
    REG_0x164   = 0x164, Register32
    ERROR_ADDR_LO   = 0x170, Register32
    ERROR_ADDR_HI   = 0x174, Register32
    REG_0x178   = 0x178, Register32     # hwrev 2 only
    REG_0x180   = irange(0x180, 4, 4), Register32
    REG_0x1a0   = irange(0x1a0, 8, 4), Register32
    ERR_SECONDARY   = irange(0x1c0, 8, 4), Register32

    # Write bits to _PROTECT to protect them.
    # They can be unprotected by writing to _UNPROTECT unless _LOCK is written.
    # If _LOCK is written, protection can be enabled but not disabled.
    REG_PROTECT         = 0x200, R_PROTECT
    REG_UNPROTECT       = 0x204, R_PROTECT
    REG_PROTECT_LOCK    = 0x208, R_PROTECT

    # Tunables touch this, can set bits FF00001F, RW
    REG_0x20c       = 0x20c, Register32

    DIAG_LOCK   = 0x210, R_DIAG_LOCK

    # All unknown, related to transaction queueing???

    # can set bits 3FFFFFFC, RW
    REG_0x218   = 0x218, Register32
    # Tunables touch this, can set bits 000F0F0F, RW
    REG_0x220   = 0x220, Register32
    # Tunables touch this, can set bits 00FFFFFF, RW
    REG_0x224   = 0x224, Register32
    # can set bits 3F3F3F3F
    TLIMIT      = 0x228, Register32
    # can set bits 07070707
    TEQRESERVE  = 0x22c, Register32
    # RO, outstanding transaction count???
    TRANS       = irange(0x230, 4, 4), Register32

    # hwrev 2 only for all of these
    REG_0x300   = 0x300, Register32
    REG_0x308   = 0x308, Register32
    REG_0x310   = 0x310, Register32
    REG_0x318   = 0x318, Register32
    REG_0x320   = 0x320, Register32
    REG_0x328   = 0x328, Register32
    REG_0x330   = 0x330, Register32
    REG_0x338   = 0x338, Register32
    REG_0x340   = 0x340, Register32
    REG_0x348   = 0x348, Register32
    REG_0x350   = 0x350, Register32
    REG_0x358   = 0x358, Register32

    # Unknown
    REG_0x400   = 0x400, Register32     # can set 00000003
    REG_0x404   = 0x404, Register32     # can set 001FFFFF
    REG_0x408   = 0x408, Register32     # can set 00FFFFFC
    REG_0x410   = 0x410, Register32     # can set 3FFFFFFC

    # These registers exist even though it's "not supported"
    TZ_CONFIG           = 0x500, Register32     # 3 bits
    TZ_SELECT           = 0x504, Register32     # 1 bit
    TZ_REGION0_START    = 0x508, Register32
    TZ_REGION0_END      = 0x510, Register32
    TZ_REGION0_OFFSET   = 0x518, Register32
    TZ_REGION1_START    = 0x520, Register32
    TZ_REGION1_END      = 0x528, Register32
    TZ_REGION1_OFFSET   = 0x530, Register32
    TZ_REGION2_START    = 0x538, Register32
    TZ_REGION2_END      = 0x540, Register32
    TZ_REGION2_OFFSET   = 0x548, Register32

    # completely guessed, unverified, can set bits 0F077077
    PERF_INTR_ENABLE    = 0x700, Register32
    PERF_INTR_STATUS    = 0x704, Register32

    PERF_UNK1   = irange(0x720, 8, 4), Register32
    PERF_UNK2   = irange(0x740, 8, 4), Register32

    PERF_TLB_MISS       = 0x760, Register32
    PERF_TLB_FILL       = 0x764, Register32
    PERF_TLB_HIT        = 0x768, Register32
    PERF_ST_MISS        = 0x770, Register32
    PERF_ST_FILL        = 0x774, Register32
    PERF_ST_HIT         = 0x778, Register32
    # hwrev 1 doesn't have these
    PERF_CTC_MISS       = 0x780, Register32
    PERF_CTC_FILL       = 0x784, Register32
    PERF_CTC_HIT        = 0x788, Register32

    UNK_TUNABLES    = irange(0x800, 256, 4), Register32

    ENABLE_STREAMS  = irange(0xc00, 8, 4), Register32
    DISABLE_STREAMS = irange(0xc20, 8, 4), Register32

    TCR             = irange(0x1000, 256, 4), R_TCR
    TTBR            = irange(0x1400, 256, 4), R_TTBR


class DART8110(Reloadable):
    PAGE_BITS = 14
    PAGE_SIZE = 1 << PAGE_BITS

    L0_OFF = 36
    L1_OFF = 25
    L2_OFF = 14

    IDX_BITS = 11
    Lx_SIZE = (1 << IDX_BITS)
    IDX_MASK = Lx_SIZE - 1

    def __init__(self, iface, regs, util=None):
        self.iface = iface
        self.regs = regs
        self.u = util
        self.pt_cache = {}

        enabled_streams = 0
        for i in range(8):
            enabled_streams |= regs.ENABLE_STREAMS[i].val << 32*i
        self.enabled_streams = enabled_streams

    @classmethod
    def from_adt(cls, u, path, instance=0, **kwargs):
        dart_addr = u.adt[path].get_reg(instance)[0]
        regs = DART8110Regs(u, dart_addr)
        dart = cls(u.iface, regs, u, **kwargs)
        return dart

    def iomap_at(self, stream, iova, addr, size):
        if size == 0:
            return

        if not (self.enabled_streams & (1 << stream)):
            self.enabled_streams |= (1 << stream)
            self.regs.ENABLE_STREAMS[stream // 32].val |= (1 << (stream % 32))

        tcr = self.regs.TCR[stream].reg

        if tcr.BYPASS_DART and not tcr.TRANSLATE_ENABLE:
            raise Exception("Stream is bypassed in DART")

        if tcr.BYPASS_DART or not tcr.TRANSLATE_ENABLE:
            raise Exception(f"Unknown DART mode {tcr}")

        if addr & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned PA {addr:#x}")

        if iova & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned IOVA {iova:#x}")

        start_page = align_down(iova, self.PAGE_SIZE)
        end = iova + size
        end_page = align_up(end, self.PAGE_SIZE)

        dirty = set()

        for page in range(start_page, end_page, self.PAGE_SIZE):
            paddr = addr + page - start_page

            ttbr = self.regs.TTBR[stream].reg
            if not ttbr.VALID:
                l1addr = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                self.pt_cache[l1addr] = [0] * self.Lx_SIZE
                ttbr.VALID = 1
                ttbr.ADDR = l1addr >> self.PAGE_BITS
                self.regs.TTBR[stream].reg = ttbr

            if tcr.FOUR_LEVELS:
                cached, l0 = self.get_pt(ttbr.ADDR << self.PAGE_BITS)
                l0idx = (page >> self.L0_OFF) & self.IDX_MASK
                l0pte = PTE(l0[l0idx])
                if not l0pte.VALID:
                    l1addr = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                    self.pt_cache[l1addr] = [0] * self.Lx_SIZE
                    l0pte = PTE(
                        OFFSET=l1addr >> self.PAGE_BITS, VALID=1)
                    l0[l0idx] = l0pte.value
                    dirty.add(ttbr.ADDR << self.PAGE_BITS)
                else:
                    l2addr = l1pte.OFFSET << self.PAGE_BITS
                l1page = l0pte.OFFSET
            else:
                l1page = ttbr.ADDR

            cached, l1 = self.get_pt(l1page << self.PAGE_BITS)
            l1idx = (page >> self.L1_OFF) & self.IDX_MASK
            l1pte = PTE(l1[l1idx])
            if not l1pte.VALID:
                l2addr = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                self.pt_cache[l2addr] = [0] * self.Lx_SIZE
                l1pte = PTE(
                    OFFSET=l2addr >> self.PAGE_BITS, VALID=1)
                l1[l1idx] = l1pte.value
                dirty.add(l1page << self.PAGE_BITS)
            else:
                l2addr = l1pte.OFFSET << self.PAGE_BITS

            dirty.add(l1pte.OFFSET << self.PAGE_BITS)
            cached, l2 = self.get_pt(l2addr)
            l2idx = (page >> self.L2_OFF) & self.IDX_MASK
            self.pt_cache[l2addr][l2idx] = PTE(
                SP_START=0, SP_END=0xfff,
                OFFSET=paddr >> self.PAGE_BITS, VALID=1).value

        for page in dirty:
            self.flush_pt(page)

    def iotranslate(self, stream, start, size):
        if size == 0:
            return []

        tcr = self.regs.TCR[stream].reg

        if tcr.BYPASS_DART and not tcr.TRANSLATE_ENABLE:
            # FIXME this may not be correct
            return [(start, size)]

        if tcr.BYPASS_DART or not tcr.TRANSLATE_ENABLE:
            raise Exception(f"Unknown DART mode {tcr}")

        if tcr.FOUR_LEVELS:
            start = start & 0xfff_ffff_ffff
        else:
            start = start & 0xfffffffff

        start_page = align_down(start, self.PAGE_SIZE)
        start_off = start - start_page
        end = start + size
        end_page = align_up(end, self.PAGE_SIZE)
        end_size = end - (end_page - self.PAGE_SIZE)

        pages = []

        for page in range(start_page, end_page, self.PAGE_SIZE):
            ttbr = self.regs.TTBR[stream].reg
            if not ttbr.VALID:
                pages.append(None)
                continue

            if tcr.FOUR_LEVELS:
                cached, l0 = self.get_pt(ttbr.ADDR << self.PAGE_BITS)
                l0pte = PTE(l0[(page >> self.L0_OFF) & self.IDX_MASK])
                if not l0pte.VALID and cached:
                    cached, l0 = self.get_pt(ttbr.ADDR << self.PAGE_BITS, uncached=True)
                    l0pte = PTE(l0[(page >> self.L0_OFF) & self.IDX_MASK])
                if not l0pte.VALID:
                    pages.append(None)
                    continue
                l1page = l0pte.OFFSET
            else:
                l1page = ttbr.ADDR

            cached, l1 = self.get_pt(l1page << self.PAGE_BITS)
            l1pte = PTE(l1[(page >> self.L1_OFF) & self.IDX_MASK])
            if not l1pte.VALID and cached:
                cached, l1 = self.get_pt(l1page << self.PAGE_BITS, uncached=True)
                l1pte = PTE(l1[(page >> self.L1_OFF) & self.IDX_MASK])
            if not l1pte.VALID:
                pages.append(None)
                continue

            cached, l2 = self.get_pt(l1pte.OFFSET << self.PAGE_BITS)
            l2pte = PTE(l2[(page >> self.L2_OFF) & self.IDX_MASK])
            if not l2pte.VALID and cached:
                cached, l2 = self.get_pt(l1pte.OFFSET << self.PAGE_BITS, uncached=True)
                l2pte = PTE(l2[(page >> self.L2_OFF) & self.IDX_MASK])
            if not l2pte.VALID:
                pages.append(None)
                continue

            pages.append(l2pte.OFFSET << self.PAGE_BITS)

        ranges = []

        for page in pages:
            if not ranges:
                ranges.append((page, self.PAGE_SIZE))
                continue
            laddr, lsize = ranges[-1]
            if ((page is None and laddr is None) or
                (page is not None and laddr == (page - lsize))):
                ranges[-1] = laddr, lsize + self.PAGE_SIZE
            else:
                ranges.append((page, self.PAGE_SIZE))

        ranges[-1] = (ranges[-1][0], ranges[-1][1] - self.PAGE_SIZE + end_size)

        if start_off:
            ranges[0] = (ranges[0][0] + start_off if ranges[0][0] else None,
                         ranges[0][1] - start_off)

        return ranges

    def get_pt(self, addr, uncached=False):
        cached = True
        if addr not in self.pt_cache or uncached:
            cached = False
            self.pt_cache[addr] = list(
                struct.unpack(f"<{self.Lx_SIZE}Q", self.iface.readmem(addr, self.PAGE_SIZE)))

        return cached, self.pt_cache[addr]

    def flush_pt(self, addr):
        assert addr in self.pt_cache
        self.iface.writemem(addr, struct.pack(f"<{self.Lx_SIZE}Q", *self.pt_cache[addr]))

    def initialize(self):
        for i in range(15):
            self.regs.TCR[i].reg = R_TCR(TRANSLATE_ENABLE=1)
        self.regs.TCR[15].reg = R_TCR(BYPASS_DART=1)

        for i in range(16):
            self.regs.TTBR[i].reg = R_TTBR(VALID = 0)

        # self.regs.ERROR.val = 0xffffffff
        # self.regs.UNK1.val = 0
        self.regs.DISABLE_STREAMS[0].val = 0xffff
        self.enabled_streams = 0

        self.invalidate_streams()

    def show_error(self):
        if self.regs.ERROR.reg.FLAG:
            print(f"ERROR: {self.regs.ERROR.reg!s}")
            print(f"ADDR: {self.regs.ERROR_ADDR_HI.val:#x}:{self.regs.ERROR_ADDR_LO.val:#x}")
            self.regs.ERROR.val = 0x80000004

    def invalidate_streams(self, streams=0xffff):
        for sid in range(256):
            if streams & (1 << sid):
                self.regs.TLB_OP.val = R_TLB_OP(STREAM=sid, OP=1)
                while self.regs.TLB_OP.reg.BUSY:
                    pass

    def invalidate_cache(self):
        self.pt_cache = {}

    def dump_table2(self, base, l1_addr, indent=""):

        def print_block(base, pte, start, last):
            pgcount = last - start
            pte.OFFSET -= pgcount
            print(indent + "    page (%4d): %09x ... %09x -> %016x [%d%d%d%d]" % (
                    start, base + start*0x4000, base + (start+1)*0x4000,
                    pte.OFFSET << self.PAGE_BITS,
                    pte.RDPROT, pte.WRPROT, pte.UNCACHABLE, pte.VALID))
            if start < last:
                print(indent + "     ==> (%4d):           ... %09x -> %016x size: %08x" % (
                    last, base + (last+1)*0x4000,
                    (pte.OFFSET + pgcount - 1) << self.PAGE_BITS, pgcount << self.PAGE_BITS))

        cached, tbl = self.get_pt(l1_addr)

        unmapped = False
        start = 0
        next_pte = PTE(VALID=0)

        for i, pte in enumerate(tbl):
            pte = PTE(pte)
            if not pte.VALID:
                if not unmapped:
                    if next_pte.VALID:
                        print_block(base, next_pte, start, i)
                    print(indent + "  ...")
                    unmapped = True
                    next_pte = pte
                continue

            unmapped = False

            if int(pte) != int(next_pte):
                if next_pte.VALID:
                    print_block(base, next_pte, start, i)
                start = i

            next_pte = pte
            next_pte.OFFSET += 1

        if next_pte.VALID:
            print_block(base, next_pte, start, 2048)

    def dump_table(self, base, l1_addr, indent="", four_levels=False):
        cached, tbl = self.get_pt(l1_addr)

        unmapped = False
        for i, pte in enumerate(tbl):
            pte = PTE(pte)
            if not pte.VALID:
                if not unmapped:
                    print("  ...")
                    unmapped = True
                continue

            unmapped = False

            if four_levels:
                off = self.L0_OFF
            else:
                off = self.L1_OFF
            print(indent + "  table (%d): %09x ... %09x -> %016x [%d%d%d%d]" % (
                i, base + (i << off), base + ((i+1) << off),
                pte.OFFSET << self.PAGE_BITS,
                pte.RDPROT, pte.WRPROT, pte.UNCACHABLE, pte.VALID))
            if four_levels:
                self.dump_table(base + (i << off), pte.OFFSET << self.PAGE_BITS,
                                indent=indent+"  ")
            else:
                self.dump_table2(base + (i << off), pte.OFFSET << self.PAGE_BITS)

    def dump_ttbr(self, ttbr, four_levels=False):
        if not ttbr.VALID:
            return

        l1_addr = (ttbr.ADDR) << self.PAGE_BITS
        print("  TTBR: %011x" % (l1_addr))

        self.dump_table(0, l1_addr, four_levels=four_levels)

    def dump_device(self, idx):
        tcr = self.regs.TCR[idx].reg
        ttbr = self.regs.TTBR[idx]
        print(f"dev {idx:02x}: TCR={tcr!s} TTBR = {ttbr!s}")

        if tcr.TRANSLATE_ENABLE and tcr.BYPASS_DART:
            print("  mode: INVALID")
        elif tcr.TRANSLATE_ENABLE:
            print("  mode: TRANSLATE")

            self.dump_ttbr(ttbr.reg, four_levels=tcr.FOUR_LEVELS)
        elif tcr.BYPASS_DART:
            print("  mode: BYPASS")
        else:
            print("  mode: UNKNOWN")

    def dump_params(self):
        print(self.regs.PARAMS_0.reg)
        print(self.regs.PARAMS_4.reg)
        print(self.regs.PARAMS_8.reg)
        print(self.regs.PARAMS_C.reg)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from construct import *

from m1n1.setup import *
from m1n1.hw.dart import DART
from m1n1.fw.aop.client import AOPClient
from m1n1.fw.aop.ipc import *

# aop nodes have no clocks described in adt for j293. it does it itself
p.pmgr_adt_clocks_enable("/arm-io/aop")
p.pmgr_adt_clocks_enable("/arm-io/dart-aop")

dart = DART.from_adt(u, "/arm-io/dart-aop",
                     iova_range=(u.adt["/arm-io/dart-aop"].vm_base, 0x1000000000))
dart.initialize()

aop = AOPClient(u, "/arm-io/aop", dart)
aop.update_bootargs({
    'p0CE': 0x20000,
    'laCn': 0x0,
    'tPOA': 0x1,
    "gila": 0x80,
})
aop.verbose = 4

p.dapf_init_all()
aop.asc.OUTBOX_CTRL.val = 0x20001 # (FIFOCNT=0x0, OVERFLOW=0, EMPTY=1, FULL=0, RPTR=0x0, WPTR=0x0, ENABLE=1)

aop.start()
for epno in [0x20, 0x21, 0x22, 0x24, 0x25, 0x26, 0x27, 0x28]:
    aop.start_ep(epno)
aop.work_for(0.3)
alsep = aop.als
alsep.VERBOSE = True

def start():
    #ret = alsep.roundtrip(GetProperty(key=AOPPropKey.MANUFACTURER))
    #print(ret.rets.value.decode("ascii")) # FireFish2
    ret = alsep.send_notify(GetProperty(key=AOPPropKey.MANUFACTURER))
    ret = alsep.send_notify(ALSSetPropertyVerbosity(level=0xffffffff)) # retcode: 0
    # [syslog] * [ALSAOPDriver.cpp:267]setProperty Setting log level = -1

    dump = """
00000000  09 00 00 10 00 00 00 00  02 24 00 00 00 00 00 00  |.........$......|
00000010  00 00 00 00 00 00 00 00  99 00 00 00 02 10 04 00  |................|
00000020  24 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |$...............|
00000030  00 00 00 00 0b 00 00 00  01 02 00 00 00 ad ab 0a  |................|
00000040  00 01 91 00 9a 59 fa 00  64 00 64 00 08 08 08 04  |.....Y..d.d.....|
00000050  09 63 eb ff ff d4 0e 00  00 c7 ce ff ff 23 2e 00  |.c...........#..|
00000060  00 a2 08 cd 14 d6 5f 77  05 00 00 00 00 00 02 00  |......_w........|
00000070  80 aa 7e 53 7f 10 84 d9  7e 68 7f 68 7f b3 7f f1  |..~S....~h.h....|
00000080  7d 00 00 00 00 00 03 5b  7e 56 7e 00 80 ca 83 e0  |}......[~V~.....|
00000090  7e 79 7f 00 80 8d 80 f1  7f 00 00 00 00 00 02 f3  |~y..............|
000000a0  7c 00 80 38 7f 88 83 0b  7f 05 7f b9 7f b9 7f 0f  ||..8............|
000000b0  7e 00 00 00 00 00 06 75  7e 38 7f 69 7f bf 83 f2  |~......u~8.i....|
000000c0  7e 16 7f 4e 7f 33 7f 1f  7d                       |~..N.3..}       |
    """
    ret = alsep.send_notify(ALSSetPropertyCalibration(value=chexundump(dump)[0x38:]))
    # [syslog] * [ALSCT720.cpp:1138]loadALSCalibrationData: ALS sensor is calibrated
    # retcode: 0xe00002bc

    ret = alsep.send_notify(ALSSetPropertyInterval(interval=200000))
    # [syslog] * [ALSCT720.cpp:690]configureSensor - Configure sensor with gain 8 and integration time 197380
    # [syslog] * [ALSCT720.cpp:708]setSensorEnabled - Enabling the sensor.

    if 0: # test that on/off works
        aop.work_for(0.5)
        ret = alsep.send_notify(ALSSetPropertyInterval(interval=0))
        # [syslog] * [ALSCT720.cpp:598]setProperty: set report interval 0, _wakeHintMode = 0
        # [syslog] * [ALSCT720.cpp:735]setSensorEnabled - Disabling the sensor.
        ret = alsep.send_notify(ALSSetPropertyInterval(interval=200000))

    while True:
        aop.work()

try:
    start()
except KeyboardInterrupt:
    pass



# SPDX-License-Identifier: MIT
from pathlib import Path
import os

class KernelRegmapAccessor:
    def __init__(self, name):
        self.path = self._find_path(name)
        self.read_ranges()
        self.read_linelen()

    @classmethod
    def _find_path(cls, name):
        basedir = Path("/sys/kernel/debug/regmap")

        if (path := Path(name)).exists():
            return path
        elif (path := basedir.joinpath(name)).exists():
            return path
        elif name in (available := cls._list_regmaps(basedir)):
            return available[name]
        else:
            raise ValueError(f"kernel regmap not found: {name}")

    @classmethod
    def _list_regmaps(cls, basedir):
        return {
            p.joinpath("name").open("rb").read().strip().decode(): p
            for p in basedir.iterdir() if p.is_dir()
        }

    def open_node(self, name, mode="rb", **kwargs):
        return self.path.joinpath(name).open(mode, **kwargs)

    def read_ranges(self):
        with self.open_node("range") as f:
            self.ranges = [
                range(int(a, 16), int(b, 16) + 1)
                for a, b in (l.strip().split(b"-") for l in f)
            ]

    def read_linelen(self):
        with self.open_node("registers", buffering=0) as f:
            l = f.read(64).split(b"\n")[0]
            valstr = l.split(b":")[1].strip()
            self.linelen = len(l) + 1
            self.working_width = len(valstr) * 4

    def _find_off(self, reg):
        off = 0
        for r in self.ranges:
            if reg >= r.stop:
                off += r.stop - r.start
            else:
                off += reg - r.start
                break
        if reg not in r:
            raise ValueError(f"register {reg:04x} out of range")
        return off * self.linelen

    def _read(self, reg, width=None):
        assert width == self.working_width
        with self.open_node("registers", buffering=0) as f:
            f.seek(self._find_off(reg))
            l = f.read(self.linelen)
            regstr, valstr = l.split(b":")
            assert int(regstr, 16) == reg
            return int(valstr, 16)

    def read(self, reg, width=None):
        assert width % self.working_width == 0
        ret = 0
        for off in range(0, width // 8, self.working_width // 8):
            ret |= self._read(reg + off, self.working_width) << (8 * off)
        return ret

    def _write(self, reg, val, width=None):
        assert width == self.working_width
        with self.open_node("registers", mode="wb") as f:
            f.write(f"{reg:x} {val:x}".encode())

    def write(self, reg, val, width=None):
        assert width % self.working_width == 0
        for off in range(0, width // 8, self.working_width // 8):
            self._write(reg + off, val >> (8 * off), self.working_width)

def require_debugfs():
    if os.path.ismount("/sys/kernel/debug"):
        return
    os.system("mount -t debugfs none /sys/kernel/debug")

if __name__ == "__main__":
    require_debugfs()
    from m1n1.hw.codecs import TAS5770Regs
    tas = TAS5770Regs(KernelRegmapAccessor("tas2770"), 0)
    import code
    code.interact(local=locals())



# SPDX-License-Identifier: MIT

from ..hv import TraceMode
from ..utils import *
from ..hw import i2c
from . import ADTDevTracer

class I2CTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.ASYNC
    REGMAPS = [i2c.I2CRegs]
    NAMES = ["i2c"]

    def __init__(self, hv, devpath, verbose=False):
        super().__init__(hv, devpath, verbose=verbose)
        self.default_dev = I2CDevTracer()
        self.default_dev.i2c_tracer = self

    def init_state(self):
        self.state.txn = []
        self.state.devices = {}

    def w_MTXFIFO(self, mtxfifo):
        if self.state.txn is None:
            self.state.txn = []

        d = mtxfifo.DATA
        if mtxfifo.START:
            self.state.txn += ["S"]
        if mtxfifo.READ:
            self.state.txn += [None] * d
        else:
            self.state.txn.append(d)
        
        if mtxfifo.STOP:
            self.state.txn.append("P")
            self.flush_txn()

    def r_MRXFIFO(self, mrxfifo):
        if mrxfifo.EMPTY:
            self.log(f"Read while FIFO empty")
            return

        if not self.state.txn:
            self.log(f"Stray read: {mrxfifo}")
            return

        try:
            pos = self.state.txn.index(None)
            self.state.txn[pos] = mrxfifo.DATA
        except ValueError:
            self.log(f"Stray read: {mrxfifo}")

        self.flush_txn()

    def flush_txn(self):
        if not self.state.txn:
            return

        if self.state.txn[-1] != "P":
            return

        if not any(i is None for i in self.state.txn):
            self.handle_txn(self.state.txn)
            self.state.txn = None

    def handle_txn(self, txn):
        st = False
        dev = self.default_dev
        read = False
        for i in txn:
            if i == "S":
                st = True
                continue
            if st:
                addr = i >> 1
                dev = self.state.devices.get(addr, self.default_dev)
                read = bool(i & 1)
                dev.start(addr, read=read)
            elif i == "P":
                dev.stop()
            elif read:
                dev.read(i)
            else:
                dev.write(i)
            st = False

    def add_device(self, addr, device):
        device.hv = self.hv
        device.i2c_tracer = self
        self.state.devices[addr] = device

class I2CDevTracer(Reloadable):
    def __init__(self, addr=None, name=None, verbose=True):
        self.addr = addr
        self.name = name
        self.verbose = verbose
        self.txn = []

    def log(self, msg, *args, **kwargs):
        if self.name:
            msg = f"[{self.name}] {msg}"
        self.i2c_tracer.log(msg, *args, **kwargs)

    def start(self, addr, read):
        self.txn.append("S")
        if read:
            self.txn.append(f"{addr:02x}.r")
        else:
            self.txn.append(f"{addr:02x}.w")

    def stop(self):
        self.txn.append("P")
        if self.verbose:
            self.log(f"Txn: {' '.join(self.txn)}")
        self.txn = []

    def read(self, data):
        self.txn.append(f"{data:02x}")

    def write(self, data):
        self.txn.append(f"{data:02x}")

class I2CRegCache:
    def __init__(self):
        self.cache = {}

    def update(self, addr, data):
        self.cache[addr] = data

    def read(self, addr, width):
        data = self.cache.get(addr, None)
        if data is None:
            print(f"I2CRegCache: no cache for {addr:#x}")
        return data

    def write(self, addr, data, width):
        raise NotImplementedError("No write on I2CRegCache")

class I2CRegMapTracer(I2CDevTracer):
    REGMAP = RegMap
    ADDRESSING = (0, 1)

    def __init__(self, verbose=False, **kwargs):
        super().__init__(verbose=verbose, **kwargs)
        self._cache = I2CRegCache()
        self.regmap = self.REGMAP(self._cache, 0)
        self.page = 0x0
        self.reg = None
        self.regbytes = []

        self.npagebytes, nimmbytes = self.ADDRESSING
        self.pageshift = 8 * nimmbytes
        self.paged = self.npagebytes != 0

    def _reloadme(self):
        self.regmap._reloadme()
        return super()._reloadme()

    def start(self, addr, read):
        if not read:
            self.reg = None
            self.regbytes = []
        super().start(addr, read)

    def stop(self):
        super().stop()

    def handle_addressing(self, data):
        if self.reg is not None:
            return False

        self.regbytes.append(data)
        if len(self.regbytes)*8 >= self.pageshift:
            immediate = int.from_bytes(bytes(self.regbytes),
                                    byteorder="big")
            self.reg = self.page << self.pageshift | immediate
        return True

    @property
    def reg_imm(self):
        '''Returns the 'immediate' part of current register address'''
        return self.reg & ~(~0 << self.pageshift)

    def handle_page_register(self, data):
        if not self.paged:
            return False

        if self.reg_imm >= self.npagebytes:
            return False

        shift = 8 * self.reg_imm
        self.page &= ~(0xff << shift)
        self.page |= data << shift
        return True

    def write(self, data):
        if self.handle_addressing(data):
            return
        elif self.handle_page_register(data):
            pass
        else:
            self.regwrite(self.reg, data)

        self.reg += 1
        super().write(data)

    def read(self, data):
        if self.reg_imm >= self.npagebytes:
            self.regread(self.reg, data)
        self.reg += 1
        super().read(data)

    def regwrite(self, reg, val):
        self.regevent(reg, val, False)

    def regread(self, reg, val):
        self.regevent(reg, val, True)

    def regevent(self, reg, val, read):
        self._cache.update(reg, val)
        r, index, rcls = self.regmap.lookup_addr(reg)
        val = rcls(val) if rcls is not None else f"{val:#x}"
        regname = self.regmap.get_name(reg) if r else f"{reg:#x}"
        t = "R" if read else "W"
        self.log(f"REG: {t.upper()}.8  {regname} = {val!s}")



# SPDX-License-Identifier: MIT
import pprint
import struct, functools, time
from dataclasses import dataclass
from enum import IntEnum

from construct.lib import hexundump

from ..asc.base import *
from ...utils import *

from . import ipc
from .dcpep import CallContext

## DCP API manager

class DCPBaseManager:
    def __init__(self, dcpep):
        self.dcpep = dcpep
        self.dcp = dcpep.asc
        dcpep.mgr = self

        self.name_map = {}
        self.tag_map = {}

        self.in_callback = 0

        for k, (cls, v) in ipc.ALL_METHODS.items():
            self.name_map[v.name] = k, v
            self.tag_map[k] = v

    def handle_cb(self, state):
        method = self.tag_map.get(state.tag, None)
        if method is None:
            raise Exception(f"Unknown callback {state.tag}")

        func = getattr(self, method.name, None)

        if func is None:
            raise Exception(f"Unimplemented callback {method!s} [{state.tag}]")

        self.in_callback += 1
        try:
            retval = method.callback(func, state.in_data)
        except Exception as e:
            print(f"Exception in callback {method.name}")
            raise
        self.in_callback -= 1
        return retval

    def __getattr__(self, attr):
        tag, method = self.name_map.get(attr, (None, None))
        if method is None or tag.startswith("D"):
            raise AttributeError(f"Unknown method {attr}")

        out_len = method.out_struct.sizeof()
        if self.in_callback:
            ctx = CallContext.CB
        else:
            ctx = CallContext.CMD
        rpc = functools.partial(self.dcpep.ch_cmd.call, ctx, tag, out_len=out_len)
        return functools.partial(method.call, rpc)

class DCPManager(DCPBaseManager):
    def __init__(self, dcpep, compatible='t8103'):
        super().__init__(dcpep)

        self.iomfb_prop = {}
        self.dcpav_prop = {}
        self.service_prop = {}
        self.pr_prop = {}

        self.swaps = 0
        self.frame = 0

        self.mapid = 0
        self.bufs = {}

        self.compatible = compatible

    ## IOMobileFramebufferAP methods

    def find_swap_function_gated(self):
        pass

    def create_provider_service(self):
        return True

    def create_product_service(self):
        return True

    def create_PMU_service(self):
        return True

    def create_iomfb_service(self):
        return True

    def create_backlight_service(self):
        return False

    def setProperty(self, key, value):
        self.iomfb_prop[key] = value
        print(f"setProperty({key} = {value!r})")
        return True

    setProperty_dict = setProperty_int = setProperty_bool = setProperty_str = setProperty

    def swap_complete_ap_gated(self, swap_id, unkBool, swap_data, swap_info, unkUint):
        swap_data_ptr = "NULL" if swap_data is None else "..."
        print(f"swap_complete_ap_gated({swap_id}, {unkBool}, {swap_data_ptr}, ..., {unkUint}")
        if swap_data is not None:
            chexdump(swap_data)
        chexdump(swap_info)
        self.swaps += 1
        self.frame = swap_id

    def swap_complete_intent_gated(self, swap_id, unkB, unkInt, width, height):
        print(f"swap_complete_intent_gated({swap_id}, {unkB}, {unkInt}, {width}, {height}")
        self.swaps += 1
        self.frame = swap_id

    def enable_backlight_message_ap_gated(self, unkB):
        print(f"enable_backlight_message_ap_gated({unkB})")

    # wrapper for set_digital_out_mode to print information on the setted modes
    def SetDigitalOutMode(self, color_id, timing_id):
        color_mode = [x for x in self.dcpav_prop['ColorElements'] if x['ID'] == color_id][0]
        timing_mode = [x for x in self.dcpav_prop['TimingElements'] if x['ID'] == timing_id][0]
        pprint.pprint(color_mode)
        pprint.pprint(timing_mode)
        self.set_digital_out_mode(color_id, timing_id)

    ## UPPipeAP_H13P methods

    def did_boot_signal(self):
        return True

    def did_power_on_signal(self):
        return True

    def will_power_off_signal(self):
        return

    def rt_bandwidth_setup_ap(self, config):
        print("rt_bandwidth_setup_ap(...)")
        if self.compatible == 't8103':
            config.val = {
                "reg1": 0x23b738014, # reg[5] in disp0/dispext0, plus 0x14 - part of pmgr
                "reg2": 0x23bc3c000, # reg[6] in disp0/dispext0 - part of pmp/pmgr
                "bit": 2,
            }
        elif self.compatible == 't600x':
            config.val = {
                "reg1": 0x28e3d0000 + 0x988, # reg[4] in disp0/dispext0, plus 0x988
                "reg2": 0x0,
                "bit": 0,
            }
        else:
            raise ValueError(self.compatible)

    ## UnifiedPipeline2 methods

    def match_pmu_service(self):
        pass

    def set_number_property(self, key, value):
        pass

    def create_provider_service(self):
        return True

    def is_dark_boot(self):
        return False

    def read_edt_data(self, key, count, value):
        return False

    def UNK_get_some_field(self):
        return 0

    def start_hardware_boot(self):
        self.set_create_DFB()
        self.do_create_default_frame_buffer()
        self.setup_video_limits()
        self.flush_supportsPower(True)
        self.late_init_signal()
        self.setDisplayRefreshProperties()
        return True

    def setDCPAVPropStart(self, length):
        print(f"setDCPAVPropStart({length:#x})")
        self.dcpav_prop_len = length - 1 # off by one?
        self.dcpav_prop_off = 0
        self.dcpav_prop_data = []
        return True

    def setDCPAVPropChunk(self, data, offset, length):
        print(f"setDCPAVPropChunk(..., {offset:#x}, {length:#x})")
        assert offset == self.dcpav_prop_off
        self.dcpav_prop_data.append(data)
        self.dcpav_prop_off += len(data)
        return True

    def setDCPAVPropEnd(self, key):
        print(f"setDCPAVPropEnd({key!r})")
        blob = b"".join(self.dcpav_prop_data)
        assert self.dcpav_prop_len == len(blob)
        self.dcpav_prop[key] = ipc.OSSerialize().parse(blob)
        self.dcpav_prop_data = self.dcpav_prop_len = self.dcpav_prop_off = None
        #pprint.pprint(self.dcpav_prop[key])
        return True

    def set_boolean_property(self, key, value):
        print(f"set {key!r} = {value}")

    def removeProperty(self, key):
        print(f"removeProperty({key!r})")

    def powerstate_notify(self, unk1, unk2):
        print(f"powerstate_notify({unk1}, {unk2})")

    def create_default_fb_surface(self, width, height):
        print(f"create_default_fb_surface({width}x{height})")
        return True

    def powerUpDART(self, unk):
        print(f"powerUpDART({unk})")
        return 0

    def hotPlug_notify_gated(self, unk):
        print(f"hotPlug_notify_gated({unk})")

    def is_waking_from_hibernate(self):
        return False

    ## UPPipe2 methods

    def match_pmu_service_2(self):
        return True

    def match_backlight_service(self):
        return True

    def get_calendar_time_ms(self):
        return time.time_ns() // 1000_000

    def update_backlight_factor_prop(self, value):
        pass

    def map_buf(self, buf, vaddr, dva, unk):
        print(f"map buf {buf}, {unk}")
        paddr, dcpdva, dvasize = self.bufs[buf]
        vaddr.val = 0
        dva.val = self.dcp.disp_dart.iomap(4, paddr, dvasize)
        print(f"mapped to dva {dva}")
        return 0

    def update_backlight_factor_prop(self, unk):
        print(f"update_backlight_factor_prop {unk}")

    ## ServiceRelay methods

    def sr_setProperty(self, obj, key, value):
        self.service_prop.setdefault(obj, {})[key] = value
        print(f"sr_setProperty({obj}/{key} = {value!r})")
        return True

    def sr_getClockFrequency(self, obj, arg):
        print(f"sr_getClockFrequency({obj}, {arg})")
        return 533333328

    sr_setProperty_dict = sr_setProperty_int = sr_setProperty_bool = sr_setProperty_str = sr_setProperty

    def sr_get_uint_prop(self, obj, key, value):
        value.val = 0
        return False

    def sr_set_uint_prop(self, obj, key, value):
        print(f"sr_set_uint_prop({obj}, {key} = {value})")

    def set_fx_prop(self, obj, key, value):
        print(f"set_fx_prop({obj}, {key} = {value})")

    def sr_mapDeviceMemoryWithIndex(self, obj, index, flags, addr, length):
        assert obj == "PROV"
        addr.val, length.val = self.dcp.u.adt["/arm-io/disp0"].get_reg(index)
        print(f"sr_mapDeviceMemoryWithIndex({obj}, {index}, {flags}, {addr.val:#x}, {length.val:#x})")
        return 0

    ## PropRelay methods

    def pr_publish(self, prop_id, value):
        self.pr_prop[prop_id] = value
        print(f"pr_publish({prop_id}, {value!r})")

    ## MemDescRelay methods:

    def allocate_buffer(self, unk0, size, unk1, paddr, dva, dvasize):
        print(f"allocate_buffer({unk0}, {size}, {unk1})")

        dvasize.val = align_up(size, 4096)
        paddr.val = self.dcp.u.memalign(0x4000, size)
        dva.val = self.dcp.dart.iomap(0, paddr.val, size)

        self.mapid += 1
        print(f"Allocating {self.mapid} as {hex(paddr.val)} / {hex(dva.val)}")

        self.bufs[self.mapid] = (paddr.val, dva.val, dvasize.val)

        return self.mapid

    def map_physical(self, paddr, size, flags, dva, dvasize):
        dvasize.val = align_up(size, 4096)
        dva.val = self.dcp.dart.iomap(0, paddr, size)
        print(f"map_physical({paddr:#x}, {size:#x}, {flags}, {dva.val:#x}, {dvasize.val:#x})")

        self.mapid += 1
        return self.mapid




#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import argparse, pathlib, time

parser = argparse.ArgumentParser(description='Mach-O loader for m1n1')
parser.add_argument('-q', '--quiet', action="store_true", help="Disable framebuffer")
parser.add_argument('-n', '--no-sepfw', action="store_true", help="Do not preserve SEPFW")
parser.add_argument('-c', '--call', action="store_true", help="Use call mode")
parser.add_argument('-r', '--raw', action="store_true", help="Image is raw")
parser.add_argument('-E', '--entry-point', action="store", type=int, help="Entry point for the raw image", default=0x800)
parser.add_argument('-x', '--xnu', action="store_true", help="Set up for chainloading XNU")
parser.add_argument('payload', type=pathlib.Path)
parser.add_argument('boot_args', default=[], nargs="*")
args = parser.parse_args()

from m1n1.setup import *
from m1n1.tgtypes import BootArgs_r1, BootArgs_r2, BootArgs_r3
from m1n1.macho import MachO
from m1n1 import asm

new_base = u.base

if args.raw:
    image = args.payload.read_bytes()
    image += b"\x00\x00\x00\x00"
    entry = new_base + args.entry_point
else:
    macho = MachO(args.payload.read_bytes())
    image = macho.prepare_image()
    image += b"\x00\x00\x00\x00"
    entry = macho.entry
    entry -= macho.vmin
    entry += new_base

if args.quiet:
    p.iodev_set_usage(IODEV.FB, 0)

sepfw_start, sepfw_length = 0, 0
preoslog_start, preoslog_size = 0, 0

if not args.no_sepfw:
    sepfw_start, sepfw_length = u.adt["chosen"]["memory-map"].SEPFW
    if hasattr(u.adt["chosen"]["memory-map"], "preoslog"):
        preoslog_start, preoslog_size = u.adt["chosen"]["memory-map"].preoslog

image_size = align(len(image))
sepfw_off = image_size
image_size += align(sepfw_length)
preoslog_off = image_size
image_size += align(preoslog_size)
bootargs_off = image_size
bootargs_size = 0x4000
image_size += bootargs_size

print(f"Total region size: 0x{image_size:x} bytes")
image_addr = u.malloc(image_size)

print(f"Loading kernel image (0x{len(image):x} bytes)...")
u.compressed_writemem(image_addr, image, True)
p.dc_cvau(image_addr, len(image))

if not args.no_sepfw:
    print(f"Copying SEPFW (0x{sepfw_length:x} bytes)...")
    p.memcpy8(image_addr + sepfw_off, sepfw_start, sepfw_length)
    print(f"Adjusting addresses in ADT...")
    u.adt["chosen"]["memory-map"].SEPFW = (new_base + sepfw_off, sepfw_length)
    u.adt["chosen"]["memory-map"].BootArgs = (new_base + bootargs_off, bootargs_size)
    if hasattr(u.adt["chosen"]["memory-map"], "preoslog"):
        p.memcpy8(image_addr + preoslog_off, preoslog_start, preoslog_size)
        u.adt["chosen"]["memory-map"].preoslog = (new_base + preoslog_off, preoslog_size)

if args.xnu:
    def remove_oslog(node):
        names = node.segment_names.split(";")
        try:
            idx = names.index("__OS_LOG")
        except ValueError:
            return
        print(f"Removing __OS_LOG from {node.name}")
        names = names[:idx] + names[idx + 1:]
        node.segment_names = ";".join(names)
        node.segment_ranges = node.segment_ranges[:idx * 32] + node.segment_ranges[32 + idx * 32: ]

    for node in u.adt["/arm-io"]:
        if hasattr(node, "segment_names"):
            remove_oslog(node)

        for nub in node:
            if hasattr(nub, "segment_names"):
                remove_oslog(nub)

print("Setting secondary CPU RVBARs...")

rvbar = entry & ~0xfff
for cpu in u.adt["cpus"]:
    if cpu.state == "running":
        continue
    addr, size = cpu.cpu_impl_reg
    print(f"  {cpu.name}: [0x{addr:x}] = 0x{rvbar:x}")
    p.write64(addr, rvbar)

u.push_adt()

print("Setting up bootargs...")
tba = u.ba.copy()

tba.top_of_kernel_data = new_base + image_size

if len(args.boot_args) > 0:
    boot_args = " ".join(args.boot_args)
    if "-v" in boot_args.split():
        tba.video.display = 0
    else:
        tba.video.display = 1
    print(f"Setting boot arguments to {boot_args!r}")
    tba.cmdline = boot_args

if args.xnu:
    # Fix virt_base, since we often install m1n1 with it set to 0 which xnu does not like
    tba.virt_base = 0xfffffe0010000000 + (tba.phys_base & (32 * 1024 * 1024 - 1))
    tba.devtree = u.ba.devtree - u.ba.virt_base + tba.virt_base

if tba.revision <= 1:
    iface.writemem(image_addr + bootargs_off, BootArgs_r1.build(tba))
elif tba.revision == 2:
    iface.writemem(image_addr + bootargs_off, BootArgs_r2.build(tba))
elif tba.revision == 3:
    iface.writemem(image_addr + bootargs_off, BootArgs_r3.build(tba))

print(f"Copying stub...")

stub = asm.ARMAsm(f"""
1:
        ldp x4, x5, [x1], #16
        stp x4, x5, [x2]
        dc cvau, x2
        ic ivau, x2
        add x2, x2, #16
        sub x3, x3, #16
        cbnz x3, 1b

        ldr x1, ={entry}
        br x1
""", image_addr + image_size)

iface.writemem(stub.addr, stub.data)
p.dc_cvau(stub.addr, stub.len)
p.ic_ivau(stub.addr, stub.len)

print(f"Entry point: 0x{entry:x}")

if args.xnu and p.display_is_external():
    if p.display_start_dcp() >= 0:
        p.display_shutdown(0)

if args.call:
    print(f"Shutting down MMU...")
    try:
        p.mmu_shutdown()
    except ProxyCommandError:
        pass
    print(f"Jumping to stub at 0x{stub.addr:x}")
    p.call(stub.addr, new_base + bootargs_off, image_addr, new_base, image_size, reboot=True)
else:
    print(f"Reloading into stub at 0x{stub.addr:x}")
    p.reload(stub.addr, new_base + bootargs_off, image_addr, new_base, image_size)

iface.nop()
print("Proxy is alive again")



# SPDX-License-Identifier: MIT
import time

from .base import *
from ...utils import *

## Management endpoint
class ManagementMessage(Register64):
    TYPE    = 59, 52

class Mgmt_Hello(ManagementMessage):
    TYPE    = 59, 52, Constant(1)
    MAX_VER = 31, 16
    MIN_VER = 15, 0

class Mgmt_HelloAck(ManagementMessage):
    TYPE    = 59, 52, Constant(2)
    MAX_VER = 31, 16
    MIN_VER = 15, 0

class Mgmt_Ping(ManagementMessage):
    TYPE    = 59, 52, Constant(3)

class Mgmt_Pong(ManagementMessage):
    TYPE    = 59, 52, Constant(4)

class Mgmt_StartEP(ManagementMessage):
    TYPE    = 59, 52, Constant(5)
    EP      = 39, 32
    FLAG    = 1, 0

class Mgmt_SetIOPPower(ManagementMessage):
    TYPE    = 59, 52, Constant(6)
    STATE   = 15, 0

class Mgmt_IOPPowerAck(ManagementMessage):
    TYPE    = 59, 52, Constant(7)
    STATE   = 15, 0

class Mgmt_EPMap(ManagementMessage):
    TYPE    = 59, 52, Constant(8)
    LAST    = 51
    BASE    = 34, 32
    BITMAP  = 31, 0

class Mgmt_EPMap_Ack(ManagementMessage):
    TYPE    = 59, 52, Constant(8)
    LAST    = 51
    BASE    = 34, 32
    MORE    = 0

class Mgmt_SetAPPower(ManagementMessage):
    TYPE    = 59, 52, Constant(0xb)
    STATE   = 15, 0

class ASCManagementEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = ManagementMessage
    SHORT = "mgmt"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.syslog_started = False
        self.iop_power_state = 0
        self.ap_power_state = 0
        self.verbose = 1

    @msg_handler(1, Mgmt_Hello)
    def Hello(self, msg):
        self.log(f"Supported versions {msg.MIN_VER} .. {msg.MAX_VER}")
        # FIXME: we pick the highest version, we should negotiate
        self.send(Mgmt_HelloAck(MIN_VER=msg.MAX_VER, MAX_VER=msg.MAX_VER))
        return True

    @msg_handler(8, Mgmt_EPMap)
    def EPMap(self, msg):
        for i in range(32):
            if msg.BITMAP & (1 << i):
                epno = 32 * msg.BASE + i
                self.asc.eps.append(epno)
                if self.verbose > 0:
                    self.log(f"Adding endpoint {epno:#x}")

        self.send(Mgmt_EPMap_Ack(BASE=msg.BASE, LAST=msg.LAST, MORE=0 if msg.LAST else 1))

        if msg.LAST:
            for ep in self.asc.eps:
                if ep == 0: continue
                if ep < 0x10:
                    self.asc.start_ep(ep)
            self.boot_done()

        return True

    @msg_handler(0xb, Mgmt_SetAPPower)
    def APPowerAck(self, msg):
        if self.verbose > 0:
            self.log(f"AP power state is now {msg.STATE:#x}")
        self.ap_power_state = msg.STATE
        return True

    @msg_handler(7, Mgmt_IOPPowerAck)
    def IOPPowerAck(self, msg):
        if self.verbose > 0:
            self.log(f"IOP power state is now {msg.STATE:#x}")
        self.iop_power_state = msg.STATE
        return True

    @msg_handler(4, Mgmt_Pong)
    def Pong(self, msg):
        return True

    def start(self):
        self.log("Starting via message")
        self.send(Mgmt_SetIOPPower(STATE=0x220))

    def wait_boot(self, timeout=None):
        if timeout is not None:
            timeout += time.time()
        while self.iop_power_state != 0x20 or self.ap_power_state != 0x20:
            self.asc.work()
            if timeout and time.time() > timeout:
                raise ASCTimeout("Boot timed out")
        self.log("Startup complete")

    def start_ep(self, epno):
        self.send(Mgmt_StartEP(EP=epno, FLAG=2))

    def stop_ep(self, epno):
        self.send(Mgmt_StartEP(EP=epno, FLAG=1))

    def boot_done(self):
        self.send(Mgmt_SetAPPower(STATE=0x20))

    def ping(self):
        self.send(Mgmt_Ping())

    def stop(self, state=0x10):
        self.log("Stopping via message")
        self.send(Mgmt_SetAPPower(STATE=0x10))
        while self.ap_power_state == 0x20:
            self.asc.work()
        self.send(Mgmt_SetIOPPower(STATE=state))
        while self.iop_power_state != state:
            self.asc.work()



# SPDX-License-Identifier: MIT

from enum import IntEnum
from m1n1.utils import *


class R_XHCI_USBCMD(Register32):
    EU3S = 11
    EWE = 10
    CRS = 9
    CSS = 8
    LHCRST = 7
    HSEE = 3
    INTE = 2
    HCRST = 1
    RS = 0


class R_XHCI_USBSTS(Register32):
    HCE = 12
    CNR = 11
    SRE = 10
    RSS = 9
    SSS = 8
    PCD = 4
    EINT = 3
    HSE = 2
    HCH = 0


class R_XHCI_CRCR_LO(Register32):
    CRP = 31, 6
    CRR = 3
    CA = 2
    CS = 1
    RCS = 0


class R_XHCI_DNCTRL(Register32):
    N0_N15 = 15, 0


class R_XHCI_DOORBELL(Register32):
    TASK_ID = 31, 16
    RSVD = 15, 8
    TARGET = 7, 0


class R_XHCI_PORTSC(Register32):
    CCS = 0
    PED = 1
    OCA = 3
    RESET = 4
    PLS = 8, 5
    PP = 9
    SPEED = 13, 10
    PIC = 15, 14
    LWS = 16
    CSC = 17
    PEC = 18
    WRC = 19
    OCC = 20
    PRC = 21
    PLC = 22
    CEC = 23
    CAS = 24
    WCE = 25
    WDE = 26
    WOE = 27
    DR = 30
    WPR = 31


class R_XHCI_PORTLI(Register32):
    ERROR_CNT = 15, 0
    RLC = 19, 16
    TLC = 23, 20
    RSV = 31, 24


class R_XHCI_IMAN(Register32):
    IP = 0
    IE = 1


class XhciRegs(RegMap):
    HCSPARAMS1 = 0x04, Register32
    HCSPARAMS2 = 0x08, Register32
    HCSPARAMS3 = 0x0C, Register32
    HCCPARAMS1 = 0x10, Register32
    DBOFF = 0x14, Register32
    RTSOFF = 0x18, Register32
    HCCPARAMS2 = 0x1C, Register32
    USBCMD = 0x20, R_XHCI_USBCMD
    USBSTS = 0x24, R_XHCI_USBSTS
    DNCTRL = 0x34, R_XHCI_DNCTRL
    CRCR_LO = 0x38, R_XHCI_CRCR_LO
    CRCR_HI = 0x3C, Register32
    DCBAAP_LO = 0x50, Register32
    DCBAAP_HI = 0x54, Register32

    PORTSC0 = 0x420, R_XHCI_PORTSC
    PORTPMSC0 = 0x424, Register32
    PORTLI0 = 0x428, R_XHCI_PORTLI
    PORTHLPMC0 = 0x42C, Register32

    PORTSC1 = 0x430, R_XHCI_PORTSC
    PORTPMSC1 = 0x434, Register32
    PORTLI1 = 0x438, R_XHCI_PORTLI
    PORTHLPMC1 = 0x43C, Register32

    MFINDEX = 0x440, Register32
    IMAN0 = 0x460 + 0x00, R_XHCI_IMAN
    IMOD0 = 0x460 + 0x04, Register32
    ERSTSZ0 = 0x460 + 0x08, Register32
    RSVD0 = 0x460 + 0x0C, Register32
    ERSTBA0 = 0x460 + 0x10, Register64
    ERDP0 = 0x460 + 0x18, Register64

    IMAN1 = 0x480 + 0x00, R_XHCI_IMAN
    IMOD1 = 0x480 + 0x04, Register32
    ERSTSZ1 = 0x480 + 0x08, Register32
    RSVD1 = 0x480 + 0x0C, Register32
    ERSTBA1 = 0x480 + 0x10, Register64
    ERDP1 = 0x480 + 0x18, Register64

    IMAN2 = 0x4A0 + 0x00, R_XHCI_IMAN
    IMOD2 = 0x4A0 + 0x04, Register32
    ERSTSZ2 = 0x4A0 + 0x08, Register32
    RSVD0 = 0x4A0 + 0x0C, Register32
    ERSTBA2 = 0x4A0 + 0x10, Register64
    ERDP2 = 0x4A0 + 0x18, Register64

    IMAN3 = 0x4C0 + 0x00, R_XHCI_IMAN
    IMOD3 = 0x4C0 + 0x04, Register32
    ERSTSZ3 = 0x4C0 + 0x08, Register32
    RSVD0 = 0x4C0 + 0x0C, Register32
    ERSTBA3 = 0x4C0 + 0x10, Register64
    ERDP3 = 0x4C0 + 0x18, Register64

    DOORBELL = irange(0x4E0, 256, 4), R_XHCI_DOORBELL


class R_GUSB3PIPECTL(Register32):
    PHYSOFTRST = 31
    U2SSINP3OK = 29
    DISRXDETINP3 = 28
    UX_EXIT_PX = 27
    REQP1P2P3 = 24
    DEPOCHANGE = 18
    SUSPHY = 17
    LFPSFILT = 9
    RX_DETOPOLL = 8


class R_GUSB2PHYCFG(Register32):
    PHYSOFTRST = 31
    U2_FREECLK_EXISTS = 30
    SUSPHY = 6
    ULPI_UTMI = 4
    ENBLSLPM = 8


class R_GCTL(Register32):
    U2RSTECN = 16
    PRTCAP = 14, 12
    CORESOFTRESET = 11
    SOFITPSYNC = 10
    SCALEDOWN = 6, 4
    DISSCRAMBLE = 3
    U2EXIT_LFPS = 2
    GBLHIBERNATIONEN = 1
    DSBLCLKGTNG = 0


class Dwc3CoreRegs(RegMap):
    GSBUSCFG0 = 0x100, Register32
    GSBUSCFG1 = 0x104, Register32
    GTXTHRCFG = 0x108, Register32
    GRXTHRCFG = 0x10C, Register32
    GCTL = 0x110, R_GCTL
    GEVTEN = 0x114, Register32
    GSTS = 0x118, Register32
    GUCTL1 = 0x11C, Register32
    GSNPSID = 0x120, Register32
    GGPIO = 0x124, Register32
    GUID = 0x128, Register32
    GUCTL = 0x12C, Register32
    GBUSERRADDR0 = 0x130, Register32
    GBUSERRADDR1 = 0x134, Register32
    GPRTBIMAP0 = 0x138, Register32
    GPRTBIMAP1 = 0x13C, Register32
    GHWPARAMS0 = 0x140, Register32
    GHWPARAMS1 = 0x144, Register32
    GHWPARAMS2 = 0x148, Register32
    GHWPARAMS3 = 0x14C, Register32
    GHWPARAMS4 = 0x150, Register32
    GHWPARAMS5 = 0x154, Register32
    GHWPARAMS6 = 0x158, Register32
    GHWPARAMS7 = 0x15C, Register32
    GDBGFIFOSPACE = 0x160, Register32
    GDBGLTSSM = 0x164, Register32
    GDBGBMU = 0x16C, Register32
    GDBGLSPMUX = 0x170, Register32
    GDBGLSP = 0x174, Register32
    GDBGEPINFO0 = 0x178, Register32
    GDBGEPINFO1 = 0x17C, Register32
    GPRTBIMAP_HS0 = 0x180, Register32
    GPRTBIMAP_HS1 = 0x184, Register32
    GPRTBIMAP_FS0 = 0x188, Register32
    GPRTBIMAP_FS1 = 0x18C, Register32
    GUCTL2 = 0x19C, Register32
    GUSB2PHYCFG = 0x200, R_GUSB2PHYCFG
    GUSB2I2CCTL = 0x240, Register32
    GUSB2PHYACC = 0x280, Register32
    GUSB3PIPECTL = 0x2C0, R_GUSB3PIPECTL
    DWC3_GHWPARAMS8 = 0x600, Register32
    DWC3_GUCTL3 = 0x60C, Register32
    DWC3_GFLADJ = 0x630, Register32
    DWC3_GHWPARAMS9 = 0x680, Register32

    DWC3_D38 = 0xd38, Register32
    DWC3_D3C = 0xd3c, Register32
    DWC3_D40 = 0xd40, Register32


class R_PIPEHANDLER_OVERRIDE(Register32):
    RXVALID = 0
    RXDETECT = 2


class E_PIPEHANDLER_MUX_MODE(IntEnum):
    USB3_PHY = 0
    UNK1 = 1
    DUMMY_PHY = 2
    UNK3 = 3


class E_PIPEHANDLER_CLK_SELECT(IntEnum):
    UNK0 = 0
    USB3_PHY = 1
    UNK2 = 2
    UNK3 = 3
    DUMMY_PHY = 4


class R_PIPEHANDLER_MUX_CTRL(Register32):
    MUX_MODE = 1, 0, E_PIPEHANDLER_MUX_MODE
    CLK_SELECT = 5, 3, E_PIPEHANDLER_CLK_SELECT


class R_PIPEHANDLER_LOCK(Register32):
    LOCK_EN = 0


class R_PIPEHANDLER_AON_GEN(Register32):
    DWC3_FORCE_CLAMP_EN = 4
    DWC3_RESET_N = 0


class R_PIPEHANDLER_NONSELECTED_OVERRIDE(Register32):
    NATIVE_POWER_DOWN = 3, 0
    NATIVE_RESET = 12
    DUMMY_PHY_EN = 15


class PipehandlerRegs(RegMap):
    PIPEHANDLER_OVERRIDE = 0x00, R_PIPEHANDLER_OVERRIDE
    PIPEHANDLER_OVERRIDE_VALUES = 0x04, R_PIPEHANDLER_OVERRIDE
    PIPEHANDLER_MUX_CTRL = 0x0C, R_PIPEHANDLER_MUX_CTRL
    PIPEHANDLER_LOCK_REQ = 0x10, R_PIPEHANDLER_LOCK
    PIPEHANDLER_LOCK_ACK = 0x14, R_PIPEHANDLER_LOCK
    PIPEHANDLER_AON_GEN = 0x1C, R_PIPEHANDLER_AON_GEN
    PIPEHANDLER_NONSELECTED_OVERRIDE = 0x20, R_PIPEHANDLER_NONSELECTED_OVERRIDE



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from construct import *

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.hw.dart import DART
from m1n1.fw.aop.client import AOPClient
from m1n1.fw.aop.ipc import *

# on the first terminal: M1N1DEVICE=/dev/ttyACM0 python3 experiments/aop_audio.py
# at that point (hpai at pw1), aop powers on admac and you should be dropped into a shell
# on a second terminal in parallel: M1N1DEVICE=/dev/ttyACM1 tools/admac_stream.py --node admac-aop-audio --channel 1 -v | xxd -g 4 -c 12 -e
# inside the shell (on the first terminal), aop_start()
# there should now be microphone samples streamed to admac
# see povik's commit #fc66046

# aop nodes have no clocks described in adt for j293. it does it itself
p.pmgr_adt_clocks_enable("/arm-io/aop")
p.pmgr_adt_clocks_enable("/arm-io/dart-aop")

# Set up a secondary proxy channel so that we can stream
# the microphone samples
p.usb_iodev_vuart_setup(p.iodev_whoami())
p.iodev_set_usage(IODEV.USB_VUART, USAGE.UARTPROXY)

pdm2 = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/audio-pdm2"]
decm = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/dc-2400000"]

pdm_config = Container(
    bytesPerSample=pdm2.bytesPerSample, # 2 ??
    clockSource=pdm2.clockSource, # 'pll '
    pdmFrequency=pdm2.pdmFrequency, # 2400000
    pdmcFrequency=pdm2.pdmcFrequency, # 24000000
    slowClockSpeed=pdm2.slowClockSpeed, # 24000000
    fastClockSpeed=pdm2.fastClockSpeed, # 24000000
    channelPolaritySelect=pdm2.channelPolaritySelect, # 256
    channelPhaseSelect=pdm2.channelPhaseSelect, # traces say 99 but device tree says 0
    unk8=0xf7600,
    unk9=0, # this should be latency (thus 15, see below) but traces say 0
    ratios=Container(
        r1=decm.ratios.r0,
        r2=decm.ratios.r1,
        r3=decm.ratios.r2,
    ),
    filterLengths=decm.filterLengths,
    coeff_bulk=120,
    coefficients=GreedyRange(Int32sl).parse(decm.coefficients),
    unk10=1,
    micTurnOnTimeMs=pdm2.micTurnOnTimeMs, # 20
    unk11=1,
    micSettleTimeMs=pdm2.micSettleTimeMs, # 50
)

decimator_config = Container(
    latency=decm.latency, # 15
    ratios=Container(
        r1=decm.ratios.r0, # 15
        r2=decm.ratios.r1, # 5
        r3=decm.ratios.r2, # 2
    ),
    filterLengths=decm.filterLengths,
    coeff_bulk=120,
    coefficients=GreedyRange(Int32sl).parse(decm.coefficients),
)

dart = DART.from_adt(u, "/arm-io/dart-aop",
                     iova_range=(u.adt["/arm-io/dart-aop"].vm_base, 0x1000000000))
dart.initialize()

aop = AOPClient(u, "/arm-io/aop", dart)
aop.update_bootargs({
    'p0CE': 0x20000,
    'laCn': 0x0,
    'tPOA': 0x1,
    "gila": 0x80,
})
aop.verbose = 4

p.dapf_init_all()
aop.asc.OUTBOX_CTRL.val = 0x20001 # (FIFOCNT=0x0, OVERFLOW=0, EMPTY=1, FULL=0, RPTR=0x0, WPTR=0x0, ENABLE=1)

# incredible power state naming scheme:
# idle: in sleep state (default at boot)
# pw1 : in active state (admac powers on) but not capturing
# pwrd: start capturing
# to start capturing, we put the 'hpai' (high power audio input?) device
# from idle -> pw1 -> pwrd state. it starts capturing at pwrd
# shutdown sequence must also be pwrd -> pw1 -> idle

def aop_start():
    aop.audio.send_notify(SetDeviceProp(
        devid=u'hpai',
        modifier=202,
        data=Container(
            devid=u'hpai',
            cookie=2,
            target_pstate=u'pwrd',
            unk2=1,
        )
    ))

def aop_stop():
    aop.audio.send_notify(SetDeviceProp(
        devid='hpai',
        modifier=202,
        data=Container(
            devid='hpai',
            cookie=3,
            target_pstate='pw1 ',
            unk2=1,
        )
    ))

    aop.audio.send_notify(SetDeviceProp(
        devid='hpai',
        modifier=202,
        data=Container(
            devid='hpai',
            cookie=4,
            target_pstate='idle',
            unk2=0,
        )
    ))

def main():
    aop.start()
    for epno in [0x20, 0x21, 0x22, 0x24, 0x25, 0x26, 0x27, 0x28]:
        aop.start_ep(epno)
    aop.work_for(0.3)
    audep = aop.audio

    audep.send_notify(AttachDevice(devid='hpai')) # high power audio input; actual mic
    audep.send_notify(AttachDevice(devid='lpai')) # low power audio input; voice trigger mic
    audep.send_notify(AttachDevice(devid='pdm0')) # leap: low-energy audio processor I think
    # [syslog] * [udioPDMNodeBase.cpp:554]PDMDev<pdm0> off ->xi0 , 0->2400000 AP state 1

    # initChannelControl (<7, 7, 1, 7>)
    audep.send_notify(SetDeviceProp(devid='lpai', modifier=301, data=Container(unk1=7, unk2=7, unk3=1, unk4=7)))
    audep.send_notify(SetDeviceProp(devid='pdm0', modifier=200, data=pdm_config))
    audep.send_notify(SetDeviceProp(devid='pdm0', modifier=210, data=decimator_config))
    ret = audep.send_roundtrip(AudioPropertyState(devid="hpai"))
    print("hpai state: %s" % (ret.state)) # idle

    audep.send_notify(SetDeviceProp(
        devid='hpai',
        modifier=202,
        data=Container(
            devid='hpai',
            cookie=1,
            target_pstate='pw1 ',
            unk2=0,
        )
    ))

    ret = audep.send_roundtrip(AudioPropertyState(devid="hpai"))
    print("hpai state: %s" % (ret.state))
    assert(ret.state == "pw1 ")

try:
    main()
    pass
except KeyboardInterrupt:
    pass

run_shell(locals(), poll_func=aop.work)



from . import Tracer, TraceMode
from ..utils import *

class R_BAR(Register32):
    BASE        = 31, 4
    PREFETCH    = 3
    ADDR64      = 2
    BELOW_1M    = 1
    SPACE       = 0

class PCICfgSpace(RegMap):
    VENDOR_ID   = 0x00, Register16
    PRODUCT_ID  = 0x02, Register16
    COMMAND     = 0x04, Register16
    STATUS      = 0x06, Register16
    REV_CLASS   = 0x08, Register32
    HDR_TYPE    = 0x0e, Register8

    BAR         = irange(0x10, 6, 4), R_BAR
    ROMADDR     = 0x30, Register32
    CAP_PTR     = 0x34, Register32

class PCIeDevTracer(Tracer):
    CFGMAP = PCICfgSpace
    BARMAPS = []
    NAMES = []
    PREFIXES = []

    def __init__(self, hv, apcie, bus, dev, fn, verbose=False):
        super().__init__(hv, verbose=verbose, ident=f"{type(self).__name__}@{apcie}/{bus:02x}:{dev:02x}.{fn:1x}")
        self.busn = bus
        self.devn = dev
        self.fn = fn
        self.ecam_off = (bus << 20) | (dev << 15) | (fn << 12)
        self.apcie = hv.adt[apcie]
        self.bars = [0] * 6
        self.bar_ranges = [None] * 6
        self.cfg_space = self.CFGMAP(hv.u, self.apcie.get_reg(0)[0] + self.ecam_off)
        self.verbose = 3

    def init_state(self):
        self.state.bars = [R_BAR(0) for i in range(6)]
        self.state.barsize = [None] * 6

    @classmethod
    def _reloadcls(cls, force=False):
        cls.BARMAPS = [i._reloadcls(force) if i else None for i in cls.BARMAPS]
        return super()._reloadcls(force)

    def r_cfg_BAR(self, val, index):
        if self.state.bars[index].BASE == 0xfffffff:
            size = (0x10000000 - val.BASE) << 4
            self.log(f"BAR{index} size = {size:#x}")
            self.state.barsize[index] = size

    def w_cfg_BAR(self, val, index):
        self.state.bars[index] = val
        self.update_tracers(val, index)

    def update_tracers(self, val = None, index = None):
        self.hv.clear_tracers(self.ident)
        ecam = self.apcie.get_reg(0)[0]
        self.trace_regmap(ecam + self.ecam_off, 0x1000, self.CFGMAP,
                          name="cfg", prefix="cfg", mode=TraceMode.WSYNC)
        i = 0
        while i < 6:
            idx = i
            if i == index:
                bar = val
            else:
                bar = self.cfg_space.BAR[i].reg
            addr = bar.BASE << 4
            if bar.ADDR64 and i != 5:
                if i + 1 == index:
                    barh = val
                else:
                    barh = self.cfg_space.BAR[i + 1].reg
                addr |= barh.value << 32
                i += 2
            else:
                i += 1

            if addr in (0, 0xfffffff0, 0xffffffff00000000, 0xfffffffffffffff0):
                continue

            size = self.state.barsize[idx]

            if not size:
                self.log(f"BAR{idx} size is unknown!")
                continue

            # Add in PCIe DT addr flags to get the correct translation
            start = self.apcie.translate(addr | (0x02 << 88))

            self.log(f"Tracing BAR{idx} : {addr:#x} -> {start:#x}..{start+size-1:#x}")
            self.bar_ranges[idx] = irange(start, size)
            self.trace_bar(idx, start, size)

    def trace_bar(self, idx, start, size):
        if idx >= len(self.BARMAPS) or (regmap := self.BARMAPS[idx]) is None:
            return
        prefix = name = None
        if idx < len(self.NAMES):
            name = self.NAMES[i]
        if idx < len(self.PREFIXES):
            prefix = self.PREFIXES[i]

        self.trace_regmap(start, size, regmap, name=name, prefix=prefix)

    def start(self):
        self.update_tracers()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.proxy import REGION_RX_EL1
from m1n1.setup import *
from m1n1 import asm

p.smp_start_secondaries()
for i in range(1, 10):
    p.mmu_init_secondary(i)

aic = u.adt["arm-io/aic"].get_reg(0)[0]

mon.add(aic, 0xc000)

hacr = 0x1f000000056c01#0#xffffffff_ffffffff

hcr = HCR(u.mrs(HCR_EL2))
hcr.TIDCP = 0
hcr.TGE = 0
hcr.AMO = 1
hcr.IMO = 1
hcr.FMO = 1
print(hcr)
u.msr(HCR_EL2, hcr.value)
u.msr(HACR_EL2, hacr)
u.inst(0xd5033fdf) # isb

AIC_NR_IRQ = aic + 0x04
AIC_RST = aic + 0x10
AIC_CFG = aic + 0x14

AIC_CFG_ENABLE = 1 << 0
AIC_CFG_PREFER_PCORES = 1 << 28

AIC_RR_DELAY = aic + 0x28
AIC_CLUSTER_ENABLE_CFG = aic + 0x30
AIC_SOME_CNT = aic + 0x3c

AIC_DELAYS = aic + 0x100

AIC_IDLE_CLUSTERS = aic + 0x340

AIC_TIMESTAMPS = 0x28e101800

AIC_IRQ_CFG = aic + 0x2000
# AIC_IRQ_ROUTE: bits 3:0 ? only 0 works.
# AIC_IRQ_CFG_DELAY: bits 7:5

AIC_SW_GEN_SET = aic + 0x6000
AIC_SW_GEN_CLR = aic + 0x6200
AIC_MASK_SET = aic + 0x6400
AIC_MASK_CLR = aic + 0x6600
AIC_HW_STATE = aic + 0x6800

AIC_IRQ_CFG_2 = aic + 0x6a00
AIC_MASK2_SET = aic + 0xae00
AIC_MASK2_CLR = aic + 0xb000
AIC_HW2_STATE = aic + 0xb200

AIC_INTERRUPT_ACK = aic + 0xc000

num_irq = p.read32(AIC_NR_IRQ) & 0xffff

code = u.malloc(0x1000)

c = asm.ARMAsm("""
write32_ts:
    isb
    mrs x2, CNTPCT_EL0
    str w1, [x0]
    mov x0, x2
    isb
    ret

en_and_spin:
    msr DAIFSet, 7
    msr DAIFClr, 2
    isb
    b 1f

en_and_spin_sec:
    msr DAIFSet, 7
    #msr DAIFClr, 2
    isb
    b 1f

1:
    cmp x0, #0
    beq 2f
    sub x0, x0, #1
    b 1b
2:
    msr DAIFSet, 7
    msr DAIFClr, 2
    isb
    ret

""", code)

iface.writemem(code, c.data)
p.dc_cvau(code, len(c.data))
p.ic_ivau(code, len(c.data))

def cpoll():
    mon.poll()
    mon.poll()

def init():
    cpoll()

    p.set32(AIC_RST, 1)
    p.set32(AIC_CFG, 1 | AIC_CFG_PREFER_PCORES)
    
    cpoll()

def test_irq_routing():
    irq = 24

    p.write32(AIC_SW_GEN_CLR, 1 << irq)
    p.write32(AIC_MASK_CLR, 1 << irq)

    cpoll()
    ts = p.call(c.write32_ts, AIC_SW_GEN_SET, 1 << irq)
    print(f"IRQ triggered at time {ts:#x}")
    p.nop()

    print("w")
    cpoll()
    cpoll()
    time.sleep(0.1)

    #p.write32(AIC_SW_GEN_CLR, 1 << irq)

    cpoll()

def get_irq_state(irq):
    v = p.read32(AIC_HW_STATE + 4* (irq//32))
    return bool(v & 1<<(irq%32))


TEST_CPU = 2

u.msr(DAIF, 0)
for i in range(1, 10):
    u.msr(DAIF, 0x140, call=lambda x, *args: p.smp_call_sync(i, x | REGION_RX_EL1, *args))
    u.msr(DAIF, 0x1c0, call=lambda x, *args: p.smp_call_sync(i, x | REGION_RX_EL1, *args))
    u.msr((3,4,15,10,4), 0, call=lambda x, *args: p.smp_call_sync(i, x | REGION_RX_EL1, *args));
    mpidr = u.mrs(MIDR_EL1, call=lambda x, *args: p.smp_call_sync(i, x | REGION_RX_EL1, args[0], args[1]));
    print(i, hex(mpidr))

init()

def sec_call(x, *args):
    return p.smp_call_sync(TEST_CPU, x | REGION_RX_EL1, *args)

def cpu_call(cpu, x, *args):
    return p.smp_call_sync(cpu, x | REGION_RX_EL1, *args)

u.msr(HCR_EL2, hcr.value, call=sec_call)
u.msr(HACR_EL2, hacr, call=sec_call)

daif = u.mrs(DAIF)
print("DAIF: %x" % daif)
daif |= 0x1c0
print("DAIF: %x" % daif)
u.msr(DAIF, daif)

p.smp_call(TEST_CPU, c.en_and_spin, 0x10000000)
test_irq_routing()
p.smp_wait(TEST_CPU)

p.smp_call(TEST_CPU, c.en_and_spin, 0x10000000)
test_irq_routing()
p.smp_wait(TEST_CPU)

print(hex(c.en_and_spin))

daif = u.mrs(DAIF)
print("DAIF: %x" % daif)
daif &= ~0x1c0
print("DAIF: %x" % daif)
u.msr(DAIF, daif)

print("DAIF: %x" % daif)
daif &= ~0x1c0
u.msr(DAIF, daif)
print("DAIF: %x" % daif)
print("ISR", hex(u.mrs(ISR_EL1)))
print("ISR #2", hex(u.mrs(ISR_EL1, call=sec_call)))
u.msr(DAIF, daif)
print("test")
# test_irq_routing()
daif = u.mrs(DAIF)
print("ISR", hex(u.mrs(ISR_EL1)))
print("ISR #2", hex(u.mrs(ISR_EL1, call=sec_call)))
print("DAIF: %x" % daif)
daif &= ~0x1c0
u.msr(DAIF, daif)
print("DAIF: %x" % daif)
print("ISR", hex(u.mrs(ISR_EL1)))
print("ISR #2", hex(u.mrs(ISR_EL1, call=sec_call)))

for i in range(0, 10):
    if i == 0:
        isr = u.mrs(ISR_EL1)
    else:
        isr = u.mrs(ISR_EL1, call=lambda x, *args: cpu_call(i, x, *args))
    print(f"{i}: {isr:#x}")

u.msr(DAIF, 0x1c0)
print("call")
p.smp_call_el1(TEST_CPU, c.en_and_spin_sec, 0x20000000)
print("test")
print(time.time())
test_irq_routing()
print(time.time())
print("wait")
p.smp_wait(TEST_CPU)
print("-")

for i in range(1, 10):
    u.msr(DAIF, 0x140, call=lambda x, *args: p.smp_call_sync(i, x | REGION_RX_EL1, *args))

for i in range(0, 10):
    if i == 0:
        isr = u.mrs(ISR_EL1)
    else:
        isr = u.mrs(ISR_EL1, call=lambda x, *args: cpu_call(i, x, *args))
    print(f"{i}: {isr:#x}")

cpoll()
p.set32(AIC_RST, 1)
cpoll()



import json, sys, re, math
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("regfile")
args = parser.parse_args()

data = json.load(open(args.regfile))

name_map = {}

for reg in data:
    name = reg['name']
    enc = reg['enc']
    name_map[f"s{enc[0]}_{enc[1]}_c{enc[2]}_c{enc[3]}_{enc[4]}"] = name

def reg_lookup(m):
    s = m.group(0)
    return name_map.get(s, s)

def hex_parse(m):
    v = int(m.group(0), 0)
    if v and (v & (v - 1)) == 0:
        bit = int(math.log2(v))
        return f"BIT({bit})"
    v ^= 0xffff_ffff_ffff_ffff
    if v and (v & (v - 1)) == 0:
        bit = int(math.log2(v))
        return f"~BIT({bit})"
    return m.group(0)

for line in sys.stdin:
    line = re.sub(r"s(\d+)_(\d+)_c(\d+)_c(\d+)_(\d+)", reg_lookup, line)
    line = re.sub(r"\b0x[0-9a-f]+\b", hex_parse, line)
    sys.stdout.write(line)



# SPDX-License-Identifier: MIT
import errno, io, os, pkgutil, re, selectors, socketserver, threading, traceback
from construct import Array, BytesInteger, Container, Int32ul, Int64ul, Struct

from ...proxy import *
from ...sysreg import *
from ...utils import *

from ..types import *

__all__ = ["GDBServer"]

class GDBServer:
    __g = Struct(
        "regs" / Array(32, Int64ul),
        "pc" / Int64ul,
        "spsr" / Int32ul,
        "q" / Array(32, BytesInteger(16, swapped=True)),
        "fpsr" / Int32ul,
        "fpcr" / Int32ul,
    )
    __separator = re.compile("[,;:]")

    def __init__(self, hv, address, log):
        self.__hc = None
        self.__hg = None
        self.__hv = hv
        self.__interrupt_eventfd = os.eventfd(0, flags=os.EFD_CLOEXEC | os.EFD_NONBLOCK)
        self.__interrupt_selector = selectors.DefaultSelector()
        self.__request = None
        self.log = log

        self.__interrupt_selector.register(self.__interrupt_eventfd, selectors.EVENT_READ)

        handle = self.__handle

        class Handler(socketserver.BaseRequestHandler):
            def handle(self):
                handle(self.request)

        self.__server = socketserver.UnixStreamServer(address, Handler, False)
        self.__thread = threading.Thread(target=self.__server.serve_forever,)

    def __add_wp(self, addr, kind, lsc):
        start = addr & 7
        if start + kind > 8:
            return b"E01"

        self.__hv.add_hw_wp(addr & ~7, ((1 << kind) - 1) << start, lsc)
        return b"OK"

    def __remove_wp(self, addr):
        self.__hv.remove_hw_wp(addr & ~7)
        return b"OK"

    def __cpu(self, cpu):
        if cpu is None:
            return

        self.__hv.cpu(cpu)

    def __stop_reply(self):
        self.__hc = None
        self.__hg = None

        prefix = b"T05thread:"

        if self.__hv.exc_reason == START.EXCEPTION_LOWER:
            if self.__hv.exc_code == EXC.SYNC:
                if self.__hv.ctx.esr.EC == ESR_EC.BKPT_LOWER:
                    prefix = b"T05hwbreak:;thread:"
                elif self.__hv.ctx.esr.EC == ESR_EC.WATCH_LOWER:
                    bas = self.__hv.get_wp_bas(self.__hv.ctx.far)
                    if not bas is None and bas != 0:
                        offset = 0
                        while (bas & (1 << offset)) == 0:
                            offset += 1
                        addr = self.__hv.ctx.far + offset
                        formatted_addr = bytes(format(addr, "x"), "utf-8")
                        prefix = b"T05watch:" + formatted_addr + b";thread:"
        elif self.__hv.exc_reason == START.HV:
            if self.__hv.exc_code == HV_EVENT.USER_INTERRUPT:
                prefix = b"T02thread:"

        return prefix + bytes(format(self.__hv.ctx.cpu_id, "x"), "utf-8") + b";"

    def __wait_shell(self):
        try:
            os.eventfd_read(self.__interrupt_eventfd)
        except BlockingIOError:
            pass

        while not self.__interrupt_eventfd in (key.fileobj for key, mask in self.__interrupt_selector.select()):
            recv = self.__request.recv(1)
            if not recv:
                break

            for byte in recv:
                if byte in b"\1\3":
                    self.__hv.interrupt()
                    break

    def __eval(self, data):
        if self.log:
            self.log(f"eval: {data}")

        if len(data) < 1:
            return b""

        if data[0] in b"?":
            return self.__stop_reply()

        if data[0] in b"c":
            if len(data) != 1:
                self.__cpu(self.__hc)
                self.__hv.ctx.elr = int(data[1:].decode(), 16)

            self.__hv.cont()
            self.__wait_shell()
            return self.__stop_reply()

        if data[0] in b"g":
            self.__cpu(self.__hg)
            g = Container()
            g.regs = self.__hv.ctx.regs.copy()
            g.regs[31] = self.__hv.ctx.sp[1]
            g.pc = self.__hv.ctx.elr
            g.spsr = self.__hv.ctx.spsr.value
            g.q = self.__hv.u.q
            g.fpsr = self.__hv.u.mrs(FPSR)
            g.fpcr = self.__hv.u.mrs(FPCR)

            return bytes(GDBServer.__g.build(g).hex(), "utf-8")

        if data[0] in b"G":
            g = GDBServer.__g.parse(bytes.fromhex(data[1:].decode()))
            self.__cpu(self.__hg)

            for index in range(31):
                self.__hv.ctx.regs[index] = g.regs[index]

            self.__hv.ctx.sp[1] = g.regs[31]
            self.__hv.ctx.elr = g.pc
            self.__hv.ctx.spsr = g.spsr.value

            q = self.__hv.u.q
            for index, value in enumerate(g.q):
                q[index] = value
            self.__hv.u.push_simd()

            self.__hv.u.msr(FPSR, g.fpsr, silent=True)
            self.__hv.u.msr(FPCR, g.fpsr, silent=True)

            return b"OK"

        if data[0] in b"H":
            if len(data) > 1:
                if data[1] in b"c":
                    cpu_id = int(data[2:].decode(), 16)
                    if cpu_id in self.__hv.started_cpus:
                        self.__hc = cpu_id
                        return b"OK"

                    return b"E01"

                if data[1] in b"g":
                    cpu_id = int(data[2:].decode(), 16)
                    if cpu_id in self.__hv.started_cpus:
                        self.__hg = cpu_id
                        return b"OK"

                    return b"E01"

            return b""

        if data[0] in b"krR":
            self.__hv.reboot()

        if data[0] in b"m":
            split = GDBServer.__separator.split(data[1:].decode(), maxsplit=1)
            fields = [int(field, 16) for field in split]
            return bytes(self.__hv.readmem(fields[0], fields[1]).hex(), "utf-8")

        if data[0] in b"M":
            split = GDBServer.__separator.split(data[1:].decode(), maxsplit=2)
            mem = bytes.fromhex(split[2])[:int(split[1], 16)]
            if self.__hv.writemem(int(split[0], 16), mem) < len(mem):
                return "E22"

            return b"OK"

        if data[0] in b"p":
            number = int(data[1:].decode(), 16)
            self.__cpu(self.__hg)
            if number < 31:
                reg = GDBServer.__g.regs.subcon.subcon.build(self.__hv.ctx.regs[number])
            elif number == 31:
                reg = GDBServer.__g.regs.subcon.subcon.build(self.__hv.ctx.sp[1])
            elif number == 32:
                reg = GDBServer.__g.pc.build(self.__hv.ctx.elr)
            elif number == 33:
                reg = GDBServer.__g.spsr.build(self.__hv.ctx.spsr.value)
            elif number < 66:
                reg = GDBServer.__g.q.subcon.subcon.build(self.__hv.u.q[number - 34])
            elif number == 66:
                reg = GDBServer.__g.fpsr.build(self.__hv.u.mrs(FPSR))
            elif number == 67:
                reg = GDBServer.__g.fpcr.build(self.__hv.u.mrs(FPCR))
            else:
                return b"E01"

            return bytes(reg.hex(), "utf-8")

        if data[0] in b"P":
            partition = data[1:].partition(b"=")
            number = int(partition[0].decode(), 16)
            reg = bytes.fromhex(partition[2].decode())
            self.__cpu(self.__hg)
            if number < 31:
                self.__hv.ctx.regs[number] = GDBServer.__g.regs.subcon.subcon.unpack(reg)
            elif number == 31:
                self.__hv.ctx.regs[1] = GDBServer.__g.regs.subcon.subcon.unpack(reg)
            elif number == 32:
                self.__hv.ctx.elr = GDBServer.__g.pc.parse(reg)
            elif number == 33:
                self.__hv.ctx.spsr.value = GDBServer.__g.spsr.parse(reg)
            elif number < 66:
                self.__hv.u.q[number - 34] = GDBServer.__g.q.subcon.subcon.parse(reg)
                self.__hv.u.push_simd()
            elif number == 66:
                self.__hv.u.msr(FPSR, GDBServer.__g.fpsr.parse(reg), silent=True)
            elif number == 67:
                self.__hv.u.msr(FPCR, GDBServer.__g.fpcr.parse(reg), silent=True)
            else:
                return b"E01"

            return b"OK"

        if data[0] in b"q":
            split = GDBServer.__separator.split(data[1:].decode(), maxsplit=1)
            if split[0] == "C":
                cpu_id = self.__hg or self.__hv.ctx.cpu_id
                return b"QC" + bytes(format(cpu_id, "x"), "utf-8")

            if split[0] == "fThreadInfo":
                cpu_ids = b",".join(bytes(format(cpu.cpu_id, "x"), "utf-8") for cpu in self.__hv.adt["cpus"])
                return b"m" + cpu_ids

            if split[0] == "sThreadInfo":
                return b"l"

            if split[0] == "Rcmd":
                self.__cpu(self.__hg)
                self.__hv.run_code(split[1])
                return b"OK"

            if split[0] == "Supported":
                return b"PacketSize=65536;qXfer:features:read+;hwbreak+"

            if split[0] == "ThreadExtraInfo":
                thread_id = int(split[1], 16)
                for node in self.__hv.adt["cpus"]:
                    if node.cpu_id == thread_id:
                        return bytes(bytes(str(node), "utf-8").hex(), "utf-8")

                return b""

            if split[0] == "Xfer":
                xfer = GDBServer.__separator.split(split[1], maxsplit=4)
                if xfer[0] == "features" and xfer[1] == "read":
                    resource = os.path.join("features", xfer[2])
                    annex = pkgutil.get_data(__name__, resource)
                    if annex is None:
                        return b"E00"

                    request_offset = int(xfer[3], 16)
                    request_len = int(xfer[4], 16)
                    read = annex[request_offset:request_offset + request_len]
                    return (b"l" if len(read) < request_len else b"m") + read

                return b""

            if split[0] == "HostInfo":
                addressing_bits = bytes(str(64 - self.__hv.pac_mask.bit_count()), "utf-8")
                return b"cputype:16777228;cpusubtype:2;endian:little;ptrsize:64;watchpoint_exceptions_received:before;addressing_bits:" + addressing_bits + b";"

            return b""

        if data[0] in b"s":
            self.__cpu(self.__hc)

            if len(data) != 1:
                self.__hv.ctx.elr = int(data[1:].decode(), 16)

            self.__hv.step()
            return self.__stop_reply()

        if data[0] in b"T":
            if int(data[1:].decode(), 16) in self.__hv.started_cpus:
                return b"OK"

            return b"E01"

        if data[0] in b"X":
            partition = data[1:].partition(b":")
            split = GDBServer.__separator.split(partition[0].decode(), maxsplit=1)
            mem = partition[2][:int(split[1], 16)]
            if self.__hv.writemem(int(split[0], 16), mem) < len(mem):
                return b"E22"

            return b"OK"

        if data[0] in b"z":
            split = GDBServer.__separator.split(data[1:].decode(), maxsplit=2)
            if split[0] == "1":
                self.__hv.remove_hw_bp(int(split[1], 16))
                return b"OK"

            if split[0] == "2":
                return self.__remove_wp(int(split[1], 16))

            if split[0] == "3":
                return self.__remove_wp(int(split[1], 16))

            if split[0] == "4":
                return self.__remove_wp(int(split[1], 16))

            return b""

        if data[0] in b"Z":
            split = GDBServer.__separator.split(data[1:].decode(), maxsplit=2)
            if split[0] == "1":
                self.__hv.add_hw_bp(int(split[1], 16))
                return b"OK"

            if split[0] == "2":
                addr = int(split[1], 16)
                kind = int(split[2], 16)
                return self.__add_wp(addr, kind, DBGWCR_LSC.S)

            if split[0] == "3":
                addr = int(split[1], 16)
                kind = int(split[2], 16)
                return self.__add_wp(addr, kind, DBGWCR_LSC.L)

            if split[0] == "4":
                addr = int(split[1], 16)
                kind = int(split[2], 16)
                return self.__add_wp(addr, kind, DBGWCR_LSC.S | DBGWCR_LSC.L)

            return b""

        return b""

    def __send(self, prefix, data):
        with io.BytesIO(prefix) as buffer:
            buffer.write(prefix)

            last = 0
            for index, byte in enumerate(data):
                if not byte in b"#$}*":
                    continue

                buffer.write(data[last:index])
                buffer.write(b"}")
                buffer.write(bytes([byte ^ 0x20]))
                last = index + 1

            buffer.write(data[last:])
            checksum = (sum(buffer.getvalue()) - sum(prefix)) % 256

            buffer.write(b"#")
            buffer.write(bytes(format(checksum, "02x"), "utf-8"))

            value = buffer.getvalue()

            if self.log:
                self.log(f"send: {value}")

            self.__request.send(value)

    def __handle(self, request):
        self.__request = request
        input_buffer = b""

        if not self.__hv.in_shell:
            self.__hv.interrupt()
            self.__wait_shell()

        self.__interrupt_selector.register(self.__request, selectors.EVENT_READ)
        try:
            while True:
                recv = self.__request.recv(65536)
                if not recv:
                    break

                input_buffer += recv

                while True:
                    dollar = input_buffer.find(b"$")
                    if dollar < 0:
                        input_buffer = b""
                        break

                    sharp = input_buffer.find(b"#", dollar)
                    if sharp < 0 or len(input_buffer) < sharp + 3:
                        input_buffer = input_buffer[dollar:]
                        break

                    input_data = input_buffer[dollar + 1:sharp]
                    input_checksum = input_buffer[sharp + 1:sharp + 3]
                    input_buffer = input_buffer[sharp + 3:]

                    try:
                        parsed_input_checksum = int(input_checksum.decode(), 16)
                    except ValueError as error:
                        print(error)
                        continue

                    if (sum(input_data) % 256) != parsed_input_checksum:
                        self.__request.send(b"-")
                        continue

                    self.__request.send(b"+")

                    with io.BytesIO() as input_decoded:
                        input_index = 0
                        input_last = 0
                        while input_index < len(input_data):
                            if input_data[input_index] == b"*":
                                input_decoded.write(input_data[input_last:input_index])
                                instance = input_decoded.getvalue()[-1]
                                input_index += 1
                                input_run_len = input_data[input_index] - 29
                                input_run = bytes([instance]) * input_run_len
                                input_decoded.write(input_run)
                                input_index += 1
                                input_last = input_index
                            elif input_data[input_index] == b"}":
                                input_decoded.write(input_data[input_last:input_index])
                                input_index += 1
                                input_decoded.write(bytes([input_data[input_index] ^ 0x20]))
                                input_index += 1
                                input_last = input_index
                            else:
                                input_index += 1

                        input_decoded.write(input_data[input_last:])

                        try:
                            output_decoded = self.__eval(input_decoded.getvalue())
                        except Exception:
                            output_decoded = b"E." + bytes(traceback.format_exc(), "utf-8")

                    self.__send(b"$", output_decoded)
        finally:
            self.__interrupt_selector.unregister(self.__request)

    def notify_in_shell(self):
        os.eventfd_write(self.__interrupt_eventfd, 1)

    def activate(self):
        try:
            self.__server.server_bind()
        except OSError as error:
            if error.errno != errno.EADDRINUSE:
                raise

            os.remove(self.__server.server_address)
            self.__server.server_bind()

        self.__server.server_activate()
        self.__thread.start()

    def shutdown(self):
        os.close(self.__interrupt_eventfd)
        self.__interrupt_selector.close()
        self.__server.shutdown()
        self.__server.server_close()
        self.__thread.join()



"""
    UAT is just regular ARMv8 pagetables, shared between the gfx-asc firmware
    and the actual AGX hardware.

    The OS doesn't have direct control over it, TTBR0 and TTBR1 entries are placed at
    gpu-region-base, one pair for each context. The firmware automatically loads TTBR0/TTBR1
    on boot and whenever the context changes.
"""


import struct
from ..fw.agx.handoff import GFXHandoff
from ..utils import *
from ..malloc import Heap
from enum import IntEnum
import traceback

__all__ = []

class MemoryAttr(IntEnum):
    # ff = Normal, Outer Writeback RW, Inner Writeback RW
    Normal = 0 # Only accessed by the gfx-asc coprocessor
    # 00 = Device nGnRnE
    Device = 1
    # f4 = Normal, Outer Writeback RW, Inner NC
    Shared = 2 # Probably Outer-shareable. Shared with either the main cpu or AGX hardware
    # 4f = Normal, Outer NC, Inner Writeback RW
    UNK3 = 3
    # 00 = Device nGnRnE
    UNK4 = 4
    # ff = Normal, Outer Writeback RW, Inner Writeback RW
    UNK5 = 5
    # 00 = Device nGnRnE
    UNK6 = 6
    # 00 = Device nGnRnE
    UNK7 = 7


class TTBR(Register64):
    ASID = 63, 48
    BADDR = 47, 1
    VALID = 0

    def valid(self):
        return self.VALID == 1

    def block(self):
        return False

    def offset(self):
        return self.BADDR << 1

    def set_offset(self, offset):
        self.BADDR = offset >> 1

    def describe(self):
        return f"{self.offset():x} [ASID={self.ASID}, VALID={self.VALID}]"

class PTE(Register64):
    OS        = 55 # Owned by host os or firmware
    UXN       = 54
    PXN       = 53
    OFFSET    = 47, 14
    nG        = 11 # global or local TLB caching
    AF        = 10
    SH        = 9, 8
    AP        = 7, 6
    AttrIndex = 4, 2
    TYPE      = 1
    VALID     = 0

    def valid(self):
        return self.VALID == 1

    def block(self):
        return self.TYPE == 0


    def offset(self):
        return self.OFFSET << 14

    def set_offset(self, offset):
        self.OFFSET = offset >> 14

    def access_fw(self, gl=False):
        if not self.OS:
            return [[
                ["--", "--", "--", "--"],
                ["--", "RW", "--", "RW"],
                ["--", "RX", "--", "--"],
                ["RX", "R-", "--", "R-"],
            ], [
                ["--", "--", "--", "RW"],
                ["--", "--", "--", "RW"],
                ["RX", "--", "--", "R-"],
                ["RX", "RW", "--", "R-"],
            ]][gl][self.AP][(self.UXN << 1) | self.PXN]
        else:
            return [
                ["--", "R-", "-?", "RW"],
                ["R-", "--", "RW", "RW"],
                ["--", "--", "--", "--"],
                ["--", "--", "--", "--"],
            ][self.AP][(self.UXN << 1) | self.PXN]

    def access_gpu(self):
        if not self.OS:
            return "--"

        return [
            ["--", "R-", "-W", "RW"],
            ["--", "--", "--", "R-"],
            ["R-", "-W", "RW", "--"],
            ["--", "--", "--", "--"],
        ][self.AP][(self.UXN << 1) | self.PXN]

    def describe(self):
        if not self.valid():
            return f"<invalid> [{int(self)}:x]"

        return (
            f"{self.offset():x} [GPU={self.access_gpu()}, EL1={self.access_fw(0)}, GL1={self.access_fw(1)}, " +
            f"perm={self.OS}{self.AP:02b}{self.UXN}{self.PXN}, " +
            f"{MemoryAttr(self.AttrIndex).name}, {['Global', 'Local'][self.nG]}, " +
            f"Owner={['FW', 'OS'][self.OS]}, AF={self.AF}, SH={self.SH}] ({self.value:#x})"
        )

class Page_PTE(PTE):
    def valid(self):
        return self.VALID == 1 and self.TYPE == 1

    def block(self):
        return True

class UatAccessor(Reloadable):
    def __init__(self, uat, ctx=0):
        self.uat = uat
        self.ctx = ctx

    def translate(self, addr, width):
        paddr, _ = self.uat.iotranslate(self.ctx, addr, width)[0]
        if paddr is None:
            raise Exception(f"UAT Failed to translate {addr:#x}")
        return paddr

    def read(self, addr, width):
        return self.uat.u.read(self.translate(addr, width), width)
    def read8(self, addr):
        return self.uat.p.read8(self.translate(addr, 1))
    def read16(self, addr):
        return self.uat.p.read16(self.translate(addr, 2))
    def read32(self, addr):
        return self.uat.p.read32(self.translate(addr, 4))
    def read64(self, addr):
        return self.uat.p.read64(self.translate(addr, 8))

    def write(self, addr, data, width):
        self.uat.u.write(self.translate(addr, width), data, width)
    def write8(self, addr, data):
        self.uat.p.write8(self.translate(addr, 1), daat)
    def write16(self, addr, data):
        self.uat.p.write6(self.translate(addr, 2), data)
    def write32(self, addr, data):
        self.uat.p.write32(self.translate(addr, 4), data)
    def write64(self, addr, data):
        self.uat.p.write64(self.translate(addr, 8), data)

class UatStream(Reloadable):
    CACHE_SIZE = 0x1000

    def __init__(self, uat, ctx, addr, recurse=True):
        self.uat = uat
        self.ctx = ctx
        self.pos = addr
        self.cache = None
        self.meta_fn = None
        self.recurse = recurse

    def to_accessor(self):
        return UatAccessor(self.uat, self.ctx)

    def read(self, size):
        assert size >= 0

        data = b""
        if self.cache:
            data = self.cache[:size]
            cached = len(self.cache)
            self.pos += min(cached, size)
            if cached > size:
                self.cache = self.cache[size:]
                return data
            self.cache = None
            if cached == size:
                return data

            size -= cached

        # align any cache overreads to the next page boundary
        remaining_in_page = self.uat.PAGE_SIZE - (self.pos % self.uat.PAGE_SIZE)
        to_cache = min(remaining_in_page, self.CACHE_SIZE)

        try:
            self.cache = self.uat.ioread(self.ctx, self.pos, max(size, to_cache))
        except:
            traceback.print_exc()
            raise
        return data + self.read(size)

    def readable(self):
        return True

    def write(self, bytes):
        self.uat.iowrite(self.ctx, self.pos, bytes)
        self.pos += len(bytes)
        self.cache = None
        return len(bytes)

    def writable(self):
        return True

    def flush(self):
        self.cache = None

    def seek(self, n, wherenc=0):
        self.cache = None
        if wherenc == 0:
            self.pos = n
        elif wherenc == 2:
            self.pos += n

    def seekable(self):
        return True

    def tell(self):
        return self.pos

    def closed(self):
        return False


class UAT(Reloadable):
    NUM_CONTEXTS = 64

    PAGE_BITS = 14
    PAGE_SIZE = 1 << PAGE_BITS

    L0_SIZE = 2
    L0_OFF = 39
    L1_SIZE = 8
    L1_OFF = 36
    L2_OFF = 25
    L3_OFF = 14

    IDX_BITS = 11
    Lx_SIZE = (1 << IDX_BITS)

    LEVELS = [
        (L0_OFF, L0_SIZE, TTBR),
        (L1_OFF, L1_SIZE, PTE),
        (L2_OFF, Lx_SIZE, PTE),
        (L3_OFF, Lx_SIZE, Page_PTE),
    ]

    def __init__(self, iface, util=None, hv=None):
        self.iface = iface
        self.u = util
        self.p = util.proxy
        self.hv = hv
        self.pt_cache = {}
        self.dirty = set()
        self.dirty_ranges = {}
        self.allocator = None
        self.ttbr = None
        self.initialized = False
        self.sgx_dev = self.u.adt["/arm-io/sgx"]
        self.shared_region = self.sgx_dev.gfx_shared_region_base
        self.gpu_region = self.sgx_dev.gpu_region_base
        self.ttbr0_base = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
        self.ttbr1_base = self.sgx_dev.gfx_shared_region_base
        self.handoff = GFXHandoff(self.u)

        self.VA_MASK = 0
        for (off, size, _) in self.LEVELS:
            self.VA_MASK |= (size - 1) << off
        self.VA_MASK |= self.PAGE_SIZE - 1


    def set_l0(self, ctx, off, base, asid=0):
        ttbr = TTBR(BADDR = base >> 1, ASID = asid, VALID=(base != 0))
        print(f"[UAT] Set L0 ctx={ctx} off={off:#x} base={base:#x} asid={asid} ({ttbr})")
        self.write_pte(self.gpu_region + ctx * 16, off, 2, ttbr)

    def ioread(self, ctx, base, size):
        if size == 0:
            return b""

        ranges = self.iotranslate(ctx, base, size)

        iova = base
        data = []
        for addr, size in ranges:
            if addr is None:
                raise Exception(f"Unmapped page at iova {ctx}:{iova:#x}")
            data.append(self.iface.readmem(addr, size))
            iova += size

        return b"".join(data)

    def iowrite(self, ctx, base, data):
        if len(data) == 0:
            return

        ranges = self.iotranslate(ctx, base, len(data))

        iova = base
        p = 0
        for addr, size in ranges:
            if addr is None:
                raise Exception(f"Unmapped page at iova {ctx}:{iova:#x}")
            self.iface.writemem(addr, data[p:p + size])
            p += size
            iova += size

    # A stream interface that can be used for random access by Construct
    def iostream(self, ctx, base, recurse=True):
        return UatStream(self, ctx, base, recurse)

    # A read/write register interface like proxy/utils objects that can be used by RegMap
    def ioaccessor(self, ctx):
        return UatAccessor(self, ctx)

    def iomap(self, ctx, addr, size, **flags):
        iova = self.allocator.malloc(size)

        self.iomap_at(ctx, iova, addr, size, **flags)
        self.flush_dirty()
        return iova

    def iomap_at(self, ctx, iova, addr, size, **flags):
        if size == 0:
            return

        if addr & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned PA {addr:#x}")

        if iova & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned IOVA {iova:#x}")

        self.init()

        map_flags = {'OS': 1, 'AttrIndex': MemoryAttr.Normal, 'VALID': 1, 'TYPE': 1, 'AP': 1, 'AF': 1, 'UXN': 1}
        map_flags.update(flags)

        start_page = align_down(iova, self.PAGE_SIZE)
        end = iova + size
        end_page = align_up(end, self.PAGE_SIZE)

        for page in range(start_page, end_page, self.PAGE_SIZE):
            table_addr = self.gpu_region + ctx * 16
            for (offset, size, ptecls) in self.LEVELS:
                if ptecls is Page_PTE:
                    pte = Page_PTE(**map_flags)
                    pte.set_offset(addr)
                    self.write_pte(table_addr, page >> offset, size, pte)
                    addr += self.PAGE_SIZE
                else:
                    pte = self.fetch_pte(table_addr, page >> offset, size, ptecls)
                    if not pte.valid():
                        table = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                        self.p.memset32(table, 0, self.PAGE_SIZE)
                        pte.set_offset(table)
                        if ptecls is not TTBR:
                            pte.VALID = 1
                            pte.TYPE = 1
                            #pte.UNK0 = 1
                        self.write_pte(table_addr, page >> offset, size, pte)
                    table_addr = pte.offset()

        self.dirty_ranges.setdefault(ctx, []).append((start_page, end_page - start_page))
        #self.flush_dirty()


    def fetch_pte(self, offset, idx, size, ptecls):
        idx = idx & (size - 1)

        cached, table = self.get_pt(offset, size=size)
        pte = ptecls(table[idx])
        if not pte.valid() and cached:
            self.flush_dirty()
            cached, table = self.get_pt(offset, size=size, uncached=True)
            pte = ptecls(table[idx])

        return pte

    def write_pte(self, offset, idx, size, pte):
        idx = idx & (size - 1)

        cached, table = self.get_pt(offset, size=size)

        table[idx] = pte.value
        self.dirty.add(offset)

    def iotranslate(self, ctx, start, size):
        if size == 0:
            return []

        start = start & self.VA_MASK

        start_page = align_down(start, self.PAGE_SIZE)
        start_off = start - start_page
        end = start + size
        end_page = align_up(end, self.PAGE_SIZE)
        end_size = end - (end_page - self.PAGE_SIZE)

        pages = []

        for page in range(start_page, end_page, self.PAGE_SIZE):
            table_addr = self.gpu_region + ctx * 16
            for (offset, size, ptecls) in self.LEVELS:
                pte = self.fetch_pte(table_addr, page >> offset, size, ptecls)
                if not pte.valid():
                    break
                table_addr = pte.offset()

            if pte.valid():
                pages.append(pte.offset())
            else:
                pages.append(None)

        ranges = []

        for page in pages:
            if not ranges:
                ranges.append((page, self.PAGE_SIZE))
                continue
            laddr, lsize = ranges[-1]
            if ((page is None and laddr is None) or
                (page is not None and laddr == (page - lsize))):
                ranges[-1] = laddr, lsize + self.PAGE_SIZE
            else:
                ranges.append((page, self.PAGE_SIZE))

        ranges[-1] = (ranges[-1][0], ranges[-1][1] - self.PAGE_SIZE + end_size)

        if start_off:
            ranges[0] = (ranges[0][0] + start_off if ranges[0][0] else None,
                         ranges[0][1] - start_off)

        return ranges

    def ioperm(self, ctx, addr):
        page = align_down(addr, self.PAGE_SIZE)

        table_addr = self.gpu_region + ctx * 16
        for (offset, size, ptecls) in self.LEVELS:
            pte = self.fetch_pte(table_addr, page >> offset, size, ptecls)
            if not pte.valid():
                break
            table_addr = pte.offset()

        return pte

    def get_pt(self, addr, size=None, uncached=False):
        if size is None:
            size = self.Lx_SIZE
        cached = True
        if addr not in self.pt_cache or uncached:
            cached = False
            if self.p.read32(addr) == 0xabad1dea:
                self.pt_cache[addr] = [0xabad1dea0000] * size
            else:
                self.pt_cache[addr] = list(
                    struct.unpack(f"<{size}Q", self.iface.readmem(addr, size * 8)))

        return cached, self.pt_cache[addr]

    def flush_pt(self, addr):
        assert addr in self.pt_cache
        table = self.pt_cache[addr]
        self.iface.writemem(addr, struct.pack(f"<{len(table)}Q", *table))
        #self.p.dc_civac(addr, 0x4000)

    def flush_dirty(self):
        inval = False

        for page in self.dirty:
            self.flush_pt(page)
            inval = True

        self.dirty.clear()

        for ctx, ranges in self.dirty_ranges.items():
            asid = ctx << 48
            self.u.inst("tlbi aside1os, x0", asid)

    def invalidate_cache(self):
        self.pt_cache = {}

    def recurse_level(self, level, base, table, page_fn=None, table_fn=None):
        def extend(addr):
            if addr >= 0x80_00000000:
                addr |= 0xf00_00000000
            return addr

        offset, size, ptecls = self.LEVELS[level]

        cached, tbl = self.get_pt(table, size)
        sparse = False
        for i, pte in enumerate(tbl):
            pte = ptecls(pte)
            if not pte.valid():
                sparse = True
                continue

            range_size = 1 << offset
            start = extend(base + i * range_size)
            end = start + range_size - 1

            if pte.block():
                if page_fn:
                    page_fn(start, end, i, pte, level, sparse=sparse)
            else:
                if table_fn:
                    table_fn(start, end, i, pte, level, sparse=sparse)
                self.recurse_level(level + 1, start, pte.offset(), page_fn, table_fn)

            sparse = False

    def foreach_page(self, ctx, page_fn):
        self.recurse_level(0, 0, self.gpu_region + ctx * 16, page_fn)

    def foreach_table(self, ctx, table_fn):
        self.recurse_level(0, 0, self.gpu_region + ctx * 16, table_fn=table_fn)

    def init(self):
        if self.initialized:
            return

        print("[UAT] Initializing...")

        # Clear out any stale kernel page tables
        self.p.memset64(self.ttbr1_base + 0x10, 0, 0x3ff0)
        self.u.inst("tlbi vmalle1os")

        self.handoff.initialize()

        with self.handoff.lock():
            print(f"[UAT] TTBR0[0] = {self.ttbr0_base:#x}")
            print(f"[UAT] TTBR1[0] = {self.ttbr1_base:#x}")
            self.set_l0(0, 0, self.ttbr0_base)
            self.set_l0(0, 1, self.ttbr1_base)
            self.flush_dirty()
            self.invalidate_cache()

        print("[UAT] Init complete")

        self.initialized = True

    def bind_context(self, ctx, ttbr0_base):
        assert ctx != 0

        with self.handoff.lock():
            self.set_l0(ctx, 0, ttbr0_base, ctx)
            self.set_l0(ctx, 1, self.ttbr1_base, ctx)
            self.flush_dirty()
            self.invalidate_cache()

    def dump(self, ctx, log=print):
        def print_fn(start, end, i, pte, level, sparse):
            type = "page" if pte.block() else "table"
            if sparse:
                log(f"{'  ' * level}...")
            log(f"{'  ' * level}{type}({i:03}): {start:011x} ... {end:011x}"
                       f" -> {pte.describe()}")

        self.recurse_level(0, 0, self.gpu_region + ctx * 16, print_fn, print_fn)

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct

from m1n1.setup import *
from m1n1 import asm
from m1n1.hw.dart import DART

if len(sys.argv) > 1:
    dart_name = sys.argv[1]
else:
    dart_name = "dart-disp0"

dart = DART.from_adt(u, "arm-io/" + dart_name)
dart.dump_all()
dart.dart.regs.dump_regs()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, argparse, os
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))
sys.path.append("/home/eileen/asahi/avd")  # git clone https://github.com/eiln/avd.git
# Decode via firmware-emulated AVD instruction stream

from m1n1.setup import *
from m1n1.utils import *
from m1n1.fw.avd import *

from avd_emu import AVDEmulator
from tools.common import ffprobe

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', type=str, required=True, help="input bitstream")
    parser.add_argument('-d', '--dir', type=str, required=True, help="path to trace dir")
    parser.add_argument('-f', '--firmware', type=str, default="data/fw.bin", help="path to fw")
    parser.add_argument('-n', '--num', type=int, default=1, help="count")
    parser.add_argument('-a', '--all', action='store_true', help="run all")
    parser.add_argument('-x', '--stfu', action='store_true')
    parser.add_argument('-p', '--poll', action='store_true', help="poll iommu space")
    parser.add_argument('--save-raw', action='store_true', help="save raw yuv")
    args = parser.parse_args()
    mode = ffprobe(args.input)

    emu = AVDEmulator(args.firmware, stfu=True)
    emu.start()
    paths = os.listdir(os.path.join(args.dir))
    paths = sorted([os.path.join(args.dir, path) for path in paths if "frame" in path])
    assert(len(paths))
    num = len(paths) if args.all else args.num
    num = min(len(paths), num)

    if   (mode == "h264"):
        from avid.h264.decoder import AVDH264Decoder
        dec = AVDH264Decoder()
    elif (mode == "h265"):
        from avid.h265.decoder import AVDH265Decoder
        dec = AVDH265Decoder()
    elif (mode == "vp09"):
        from avid.vp9.decoder import AVDVP9Decoder
        dec = AVDVP9Decoder()
    else:
        raise RuntimeError("unsupported codec")
    if (args.stfu):
        dec.stfu = True
        dec.hal.stfu = True
    units = dec.setup(args.input)

    avd = AVDDevice(u)
    if   (mode == "h264"):
        avd.decoder = AVDH264Dec(avd)
    elif (mode == "h265"):
        avd.decoder = AVDH265Dec(avd)
    elif (mode == "vp09"):
        avd.decoder = AVDVP9Dec(avd)
    else:
        raise RuntimeError("unsupported codec")
    avd.decoder.winname = args.input
    if (args.stfu):
        avd.stfu = True
    avd.boot()
    avd.ioalloc_at(0x0, 0xf000000, stream=0)
    if (args.poll):
        avd.iomon.add(0x0, 0xf000000)

    for i,unit in enumerate(units[:num]):
        print(unit)
        inst = dec.decode(unit)
        path = paths[i]
        print(path)
        inst = emu.avd_cm3_cmd_decode(path)
        avd.decoder.decode(dec.ctx, unit, inst)
        if (args.save_raw):
            y_data = avd.ioread(dec.ctx.y_addr, dec.ctx.luma_size, stream=0)
            uv_data = avd.ioread(dec.ctx.uv_addr, dec.ctx.chroma_size, stream=0)
            open("data/raw-emu/%03d.bin" % (i), "wb").write(y_data + uv_data)



# SPDX-License-Identifier: MIT
from ...utils import *

from ..asc import StandardASC
from .dcpep import DCPEndpoint
from ..afk.epic import *

class DCPClient(StandardASC):
    ENDPOINTS = {
        0x20: AFKSystemEndpoint,
        0x37: DCPEndpoint,
    }

    def __init__(self, u, asc_base, dart=None, disp_dart=None):
        super().__init__(u, asc_base, dart)
        self.disp_dart = disp_dart



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, fnmatch, signal
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.fw.smc import SMCClient, SMCError

smc_addr = u.adt["arm-io/smc"].get_reg(0)[0]
smc = SMCClient(u, smc_addr)
smc.start()
smc.start_ep(0x20)

smc.verbose = 0

smcep = smc.epmap[0x20]

count = smcep.read32b("#KEY")
print(f"Key count: {count}")

print("Scanning keys...")

pats = sys.argv[1:]

vals = {}

fmts = {
    "D?CR": "#x",
    "AC-I": "#x",
    "D?FC": "#x",
    "D?VM": lambda v: (v>>8) | ((v&0xff)<<8),
    "D?VX": lambda v: (v>>8) | ((v&0xff)<<8),
    "B0RM": lambda v: (v>>8) | ((v&0xff)<<8),
    ##"BAAC": lambda v: ((v&0xff00)>>8) | ((v&0xff)<<8),
}

smcep.write8("NTAP", 1)

for i in range(count):
    k = smcep.get_key_by_index(i)
    if not any(fnmatch.fnmatchcase(k, i) for i in pats):
        continue
    if any(fnmatch.fnmatchcase('-' + k, i) for i in pats):
        continue
    length, type, flags = smcep.get_key_info(k)
    if type in ("ch8*",  "{jst"):
        continue
    if flags & 0x80:
        try:
            val = smcep.read_type(k, length, type)
            fmt = None
            for fk, fv in fmts.items():
                if fnmatch.fnmatchcase(k, fk):
                    fmt = fv
            if fmt is None:
                fmt = lambda a: ("%.02f" % a) if isinstance(a, float) else a
            elif isinstance(fmt, str):
                def ff(fmt):
                    return lambda a: f"{a:{fmt}}"
                fmt = ff(fmt)
            vals[k] = val, length, type, fmt
            print(f"#{i}: {k} = ({type}, {flags:#x}) {fmt(val)}")
        except SMCError as e:
            print(f"#{i}: {k} = ({type}, {flags:#x}) <error {e}>")
    else:
        print(f"#{i}: {k} = ({type}, {flags:#x}) <not available>")

slots = {}

def poll():
    global cnt
    reprint = cnt % 10 == 0
    changed = set()
    for k, (oval, length, type, fmt) in vals.items():
        val = smcep.read_type(k, length, type)
        if val != oval:
            if k not in slots:
                reprint = True
            slots[k] = fmt(val)
            changed.add(k)
            vals[k] = val, length, type, fmt
    if reprint:
        print("\x1b[1;4m", end="")
        for k, v in slots.items():
            wd = len(f"{v:>8}")
            print(f"{k:>{wd}s}", end=" ")
        print("\x1b[m")
    for k, v in slots.items():
        if k in changed:
            print("\x1b[32m", end="")
        print(f"{v:>8}\x1b[m", end=" ")
    print()
    cnt += 1
    time.sleep(1)

def handle_sigint(signal=None, stack=None):
    global doshell
    doshell = True

signal.signal(signal.SIGINT, handle_sigint)

doshell = False
try:
    cnt = 0
    while True:
        poll()
        if doshell:
            run_shell(globals(), msg="Interrupted")
            doshell = False
finally:
    smc.stop()
    



# SPDX-License-Identifier: MIT
import os, sys, struct, time

from .utils import *
from . import asm
from .proxy import REGION_RX_EL1
from .sysreg import *

class GPIOLogicAnalyzer(Reloadable):
    def __init__(self, u, node=None, pins={}, regs={}, div=1, cpu=1, on_pin_change=True, on_reg_change=True):
        self.u = u
        self.p = u.proxy
        self.iface = u.iface
        self.cpu = cpu
        self.base = 0
        if node is not None:
            self.base = u.adt[node].get_reg(0)[0]
        else:
            on_pin_change=False
        self.node = node
        self.pins = pins
        self.regs = regs
        assert len(pins) <= 32
        assert div > 0
        self.div = div
        self.cbuf = self.u.malloc(0x1000)
        self.dbuf = None
        self.on_pin_change = on_pin_change
        self.on_reg_change = on_reg_change
        self.p.mmu_init_secondary(cpu)
        self.tfreq = u.mrs(CNTFRQ_EL0)

    def load_regmap(self, regmap, skip=set(), regs=set()):
        base = regmap._base
        for name, (addr, rcls) in regmap._namemap.items():
            if name not in skip and (not regs or name in regs):
                self.regs[name] = base + addr, rcls

    def start(self, ticks, bufsize=0x10000):
        self.bufsize = bufsize
        if self.dbuf:
            self.u.free(self.dbuf)
        self.dbuf = self.u.malloc(bufsize)

        text = f"""
        trace:
            mov x16, x2
            add x3, x3, x2
            add x2, x2, #4
            mov x12, #-8
            mov x10, x2
            mov x6, #-1
            mov x7, #0
            ldr x8, ={self.base}
            mrs x4, CNTPCT_EL0
            isb
        1:
            ldr w15, [x16]
            cmp w15, #1
            b.eq done
            add x4, x4, x1
        2:
            mrs x5, CNTPCT_EL0
            isb
        """
        if self.div > 1:
            text += f"""
                cmp x5, x4
                b.lo 2b
            """

        for idx, pin in enumerate(self.pins.values()):
            text += f"""
            ldr w9, [x8, #{pin * 4}]
            bfi x7, x9, #{idx}, #1
            """

        if self.on_pin_change:
            text += f"""
                cmp x7, x6
                b.eq 3f
                mov x6, x7
            """
        if self.on_reg_change:
            text += f"""
                mov x11, x2
            """

        text += f"""
            str w5, [x2], #4
            str w7, [x2], #4
        """
        if self.on_reg_change:
            text += f"""
                mov x13, #0
                add x14, x12, #8
            """

        for reg in self.regs.values():
            if isinstance(reg, tuple):
                reg = reg[0]
            text += f"""
                ldr x9, ={reg}
                ldr w9, [x9]
                str w9, [x2], #4
            """
            if self.on_reg_change:
                text += f"""
                    eor w15, w9, #1
                    cmp x14, #0
                    b.eq 4f
                    ldr w15, [x14], #4
                4:
                    eor w15, w15, w9
                    orr w13, w13, w15
                """

        if self.on_reg_change:
            text += f"""
                cmp x13, #0
                b.ne 4f
                mov x2, x11
                mov x11, x12
                b 3f
            4:
            """
        text += f"""
            mov x12, x11
            cmp x2, x3
            b.hs done
        3:
            sub x0, x0, #1
            cbnz x0, 1b
        done:
            sub x0, x2, x10
            ret
        """
        
        code = asm.ARMAsm(text, self.cbuf)
        self.iface.writemem(self.cbuf, code.data)
        self.p.dc_cvau(self.cbuf, len(code.data))
        self.p.ic_ivau(self.cbuf, len(code.data))

        self.p.write32(self.dbuf, 0)

        self.p.smp_call(self.cpu, code.trace | REGION_RX_EL1, ticks, self.div, self.dbuf, bufsize - (8 + 4 * len(self.regs)))
    
    def complete(self):
        self.p.write32(self.dbuf, 1)
        wrote = self.p.smp_wait(self.cpu)
        assert wrote <= self.bufsize
        data = self.iface.readmem(self.dbuf + 4, wrote)
        self.u.free(self.dbuf)
        self.dbuf = None
        
        stride = 2 + len(self.regs)
        
        #chexdump(data)
        
        self.data = [struct.unpack("<" + "I" * stride,
                                   data[i:i + 4 * stride])
                     for i in range(0, len(data), 4 * stride)]

    def vcd(self):
        off = self.data[0][0]
        if False: #len(self.data) > 1:
            off2 = max(0, ((self.data[1][0] - off) & 0xffffffff) - 5000)
        else:
            off2 = 0
        
        #print(off, off2)
        
        vcd = []
        vcd.append("""
$timescale 1ns $end
$scope module gpio $end
""")
        sym = 0
        keys = []
        rkeys = []
                                           
        for name in self.pins:
            keys.append(f"s{sym}")
            vcd.append(f"$var wire 1 s{sym} {name} $end\n")
            sym += 1
        for name, reg in self.regs.items():
            vcd.append(f"$var reg 32 s{sym} {name} [31:0] $end\n")
            if isinstance(reg, tuple):
                subkeys = {}
                rcls = reg[1]
                rkeys.append((f"s{sym}", rcls, subkeys))
                sym += 1
                for fname in rcls().fields.keys():
                    fdef = getattr(rcls, fname)
                    if isinstance(fdef, tuple):
                        width = fdef[0] - fdef[1] + 1
                    else:
                        width = 1
                    vcd.append(f"$var reg {width} s{sym} {name}.{fname} [{width-1}:0] $end\n")
                    subkeys[fname] = (width, f"s{sym}")
                    sym += 1
            else:
                rkeys.append((f"s{sym}", None, None))
                sym += 1
        vcd.append("""
$enddefinitions $end
$dumpvars
""")

        for v in self.data:
            ts = v[0]
            val = v[1]
            regs = v[2:]
            ts = ((ts - off) & 0xffffffff) - off2
            ns = max(0, 1000000000 * ts // self.tfreq)
            vcd.append(f"#{ns}\n")
            vcd.append("\n".join(f"{(val>>i) & 1}{k}" for i, k in enumerate(keys)) + "\n")
            for (key, rcls, subkeys), v in zip(rkeys, regs):
                vcd.append(f"b{v:032b} {key}\n")
                if rcls:
                    rval = rcls(v)
                    for field, (width, key) in subkeys.items():
                        v = getattr(rval, field)
                        vcd.append(f"b{v:0{width}b} {key}\n")
                    

        ns += ns//10
        vcd.append(f"#{ns}\n" + "\n".join(f"{(val>>i) & 1}{k}" for i, k in enumerate(keys)) + "\n")
 
        return "".join(vcd)
    
    def show(self):
        with open("/tmp/dump.vcd", "w") as fd:
            fd.write(self.vcd())
        
        gtkw = ("""
[dumpfile] "/tmp/dump.vcd"
[timestart] 0
[size] 3063 1418
[pos] -1 -1
*-17.000000 2 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1
[sst_width] 288
[signals_width] 197
[sst_expanded] 1
[sst_vpaned_height] 421
@23
""" +
        "\n".join("gpio." + k for k in self.pins) + "\n" + 
        "\n".join("gpio." + k + "[31:0]" for k in self.regs) + "\n")

        with open("/tmp/dump.gtkw", "w") as fd:
            fd.write(gtkw)

        os.system("gtkwave /tmp/dump.gtkw&")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.shell import run_shell

from m1n1.fw.ane import ANE

import numpy as np
from anect import anect_convert  # pip install anect
def f16encode(x): return np.float16(x).tobytes()
def f16decode(b): return np.frombuffer(b[:2], dtype=np.float16)[0]


ane = ANE(u)
ane.power_up()
if 1:
	rnges = [(0x26bc04000, 0x26bc28000, 'engine'),]
	mon = RegMonitor(u)
	for (start, end, name) in rnges:
		mon.add(start, end-start, name=name)
	mon.poll() # should work after ane.power_up()

# curl -LJO https://www.dropbox.com/s/lpjap6w0kdlom1h/add.hwx?dl=0
anec = anect_convert("add.hwx")
req = ane.fw.setup(anec)

x1 = f16encode(1.0)
x2 = f16encode(2.0)
ane.fw.send_src(req, x1, 0)
ane.fw.send_src(req, x2, 1)

ane.tm.enqueue_tq(req)
ane.tm.execute_tq(req)

x3 = ane.fw.read_dst(req, 0)
print("what's 1+2? = %f" % f16decode(x3[:2]))


run_shell(globals(), msg="Have fun!")




# SPDX-License-Identifier: MIT
from contextlib import contextmanager, ExitStack
import sys, pathlib, os
import subprocess
import tempfile
import bisect

from .asm import NM, LD, OBJCOPY, ARMAsm

__all__ = ["LinkedProgram"]


def tool_output_lines(progname, *args):
    with subprocess.Popen([progname.replace("%ARCH", ARMAsm.ARCH)] + list(args),
            stdout=subprocess.PIPE) as proc:
        for line in proc.stdout:
            yield line.decode("ascii")
        proc.wait()
        if proc.returncode:
            raise Exception(f"{progname} (args: {args}) exited with status {proc.returncode}")

def run_tool(progname, *args, silent=False):
    subprocess.check_call([progname.replace("%ARCH", ARMAsm.ARCH)] + list(args),
                        stdout=subprocess.DEVNULL if silent else None)


class LinkedProgram:
    SOURCE_ROOT = str(pathlib.Path(__file__).resolve().parents[2])

    def __init__(self, u, base_object=None):
        self.u = u
        self.symbols = []
        self.symaddrs = dict()
        self.base_object = base_object
        self._alloced_bases = []
        self._attrs_to_clear = []
        self._load_base_symbols()

    def _load_base_symbols(self):
        if self.base_object is None:
            suffix = "-raw" if not self.m1n1_is_macho() else ""
            self.base_object = f"build/m1n1{suffix}.elf"

        addrs = self._load_elf_symbols(self.base_object, self.u.proxy.get_base())

        # sanity check: compare the .rela table between ELF and m1n1 image on target
        rela_base = addrs["_rela_start"]
        rela_length = addrs["_rela_end"] - rela_base
        rela_target = self.u.iface.readmem(rela_base, rela_length)

        tmp = os.path.join(tempfile.mkdtemp(), "bin")
        path = os.path.join(self.SOURCE_ROOT, self.base_object)
        run_tool(OBJCOPY, "-O", "binary", path, tmp, "--only-section=.rela.dyn")
        rela_objfile = open(tmp, "rb").read()

        if rela_objfile[:len(rela_target)] != rela_target:
            raise Exception(f"Mismatch between {self.base_object} and image on target")

    def m1n1_is_macho(self):
        p = self.u.proxy
        return p.read32(p.get_base()) == 0xfeedfacf

    def _load_elf_symbols(self, relpath, offset=0,
                            objname=None, ignore=""):
        path = pathlib.Path(self.SOURCE_ROOT, relpath)
        symaddrs = dict()

        for line in tool_output_lines(NM, "-g", path):
            addr_str, t, name = line.split()
            addr = int(addr_str, 16) + offset
            if t in ignore:
                continue
            self.symbols.append((addr, name, objname))
            symaddrs[name] = addr
            if t in "T" and not hasattr(self, name):
                setattr(self, name, self._wrap_call_to(addr))
                if relpath != self.base_object:
                    self._attrs_to_clear.append(name)
        self.symbols.sort()
        return symaddrs

    def load_obj(self, objfile, base=None):
        ALLOC_SIZE = 16*4096

        if base is None:
            base = self.u.heap.memalign(0x4000, ALLOC_SIZE)
            self._alloced_bases.append(base)

        objfile = os.path.join(self.SOURCE_ROOT, objfile)
        tmp = tempfile.mkdtemp() + os.sep
        elffile = tmp + "elf"
        ld_script = tmp + "ld"
        binfile = tmp + "bin"
        with open(ld_script, "w") as f:
            f.write("SECTIONS {\n")
            f.write(f". = 0x{base:x};\n")
            f.write(".text : { *(.text .text.*) }\n")
            f.write(".data : { *(.got .data .data.* .rodata .rodata.* .bss .bss.*) }\n")
            f.write("}\n")
            for sym in self.symbols:
                f.write(f"{sym[1]} = 0x{sym[0]:x};\n")
        run_tool(LD, "-EL", "-maarch64elf", "-T", ld_script, "-o", elffile, objfile)
        run_tool(OBJCOPY, "-O", "binary", elffile, binfile)
        #run_tool("objdump", "-d", elffile)
        self._load_elf_symbols(elffile, ignore="A")
        with open(binfile, "rb") as f:
            buf = f.read()
            assert len(buf) <= ALLOC_SIZE
            self.u.iface.writemem(base, buf)
            self.u.proxy.dc_cvau(base, len(buf))
            self.u.proxy.ic_ivau(base, len(buf))

    def clear_objs(self):
        for name in self._attrs_to_clear:
            delattr(self, name)
        self._attrs_to_clear = []

        for base in self._alloced_bases:
            self.u.free(base)
        self._alloced_bases = []

        self.symbols = [(a, b, objname) for (a, b, objname) \
                        in self.symbols if objname == self.base_object]

    @contextmanager
    def _copy_args_to_target(self, args):
        heap = self.u.heap
        with ExitStack() as stack:
            args_copied = []
            for arg in args:
                if type(arg) is str:
                    arg = arg.encode("ascii")
                    # fallthrough
                if type(arg) is bytes:
                    p = stack.enter_context(heap.guarded_malloc(len(arg) + 1))
                    self.u.iface.writemem(p, arg + b"\0")
                    args_copied.append(p)
                elif type(arg) is int:
                    args_copied.append(arg)
                else:
                    raise NotImplementedError(type(arg))
            yield args_copied

    def _wrap_call_to(self, addr):
        def call_symbol(*args, call=self.u.proxy.call):
            with self._copy_args_to_target(args) as args_copied:
                return call(addr, *args_copied)
        return call_symbol

    def lookup(self, addr):
        idx = bisect.bisect_left(self.symbols, (addr + 1, "", "")) - 1
        if idx < 0 or idx >= len(self.symbols):
            return None, None
        return self.symbols[idx]

    def load_inline_c(self, source):
        tmp = tempfile.mkdtemp()
        cfile = tmp + ".c"
        objfile = tmp + ".o"
        with open(cfile, "w") as f:
            f.write(source)
        run_tool("make", "-C", self.SOURCE_ROOT, "invoke_cc",
                 f"OBJFILE={objfile}", f"CFILE={cfile}", silent=True)
        self.load_obj(objfile)


if __name__ == "__main__":
    from m1n1.setup import *
    lp = LinkedProgram(u)
    lp.debug_printf("hello from the other side! (%d)\n", 42)
    lp.load_inline_c('''
        #include "utils.h"
        int add(int a, int b) {
            debug_printf("adding %d and %d\\n", a, b);
            return a + b;
        }
    ''')
    print(f"1 + 2 = {lp.add(1, 2)}")
